# raise NotImplementedError()
# from collections import OrderedDict, deque
# import copy
# import math
# import StringIO
#
# import numpy as np  # Because the python docs don't recommend log(x, y)
#                     # https://docs.python.org/2/library/math.html#math.log10
# import hyperopt
#
# import HPOlib.format_converter.configuration_space as configuration_space_module
#
# ################################################################################
# # Read functionality
# def read(pyll_string):
#     reader = PyllReader()
#     if isinstance(pyll_string, file):
#         pyll_string = pyll_string.read()
#     elif type(pyll_string) == str:
#         pass
#     else:
#         raise NotImplementedError()
#     return reader.read(pyll_string)
#
#
# class PyllReader(object):
#     # This won't parse AutoWEKA because inside a hp.choice there are lists.
#     # This parser assumes that every choice of an hp.choice is either a
#     # Literal or a dictionaries with exactly one literal, where the literal is
#     # one value of the choice hyperparameter. Adjust
#     # autoweka/src/java/autoweka/tpe/TPEExperimentConstructor.java
#     # Instead of:
#     # hp.choice('auto_param_2', [
#     #    {'attributesearch':'NONE'},
#     #    (
#     #        hp.choice('auto_param_4', [
#     #            {
#     #                'attributesearch':'weka.attributeSelection.BestFirst',
#     #                'assearch_wasbf__00__01_0_D':hp.choice('assearch_wasbf__00__01_0_D', ['2', '1', '0', ]),
#     #                'assearch_wasbf__00__02_1_INT_N':hp.quniform('assearch_wasbf__00__02_1_INT_N', 2.0, 10.0,1.0),
#     #                'assearch_wasbf__00__03_2_S':hp.choice('assearch_wasbf__00__03_2_S', ['0', ]),
#     #            }]))])
#     # write
#     # auto_param2 = hp.choice('auto_param_2', [
#     #    {'attributesearch':'NONE'},
#     #    {'attritubesearch':hp.choice('auto_param_4', [
#     #            {
#     #                'attributesearch':'weka.attributeSelection.BestFirst',
#     #                'assearch_wasbf__00__01_0_D':hp.choice('assearch_wasbf__00__01_0_D', ['2', '1', '0', ]),
#     #                'assearch_wasbf__00__02_1_INT_N':hp.quniform('assearch_wasbf__00__02_1_INT_N', 2.0, 10.0,1.0),
#     #                'assearch_wasbf__00__03_2_S':hp.choice('assearch_wasbf__00__03_2_S', ['0', ]),
#     #            }])}])
#
#
#     def __init__(self):
#         self.hyperparameters = OrderedDict()
#         self.constants = list()
#
#     def read(self, pyll_string):
#         exec pyll_string
#         space = hyperopt.pyll.as_apply(space)   # Space is generated by exec
#
#         if space.name not in ("dict", "pos_args", "switch"):
#             raise ValueError("The configuration space must consist of a dict "
#                              "or pos_args but the "
#                              "root node but is actually %s" % space.name)
#
#         root_values = self.read_object(space, None, condition=None)
#
#         if type(root_values) == dict:
#             for key in root_values:
#                 self.hyperparameters[key] = root_values[key]
#         elif isinstance(root_values, configuration_space_module
#                 .CategoricalHyperparameter):
#             self.hyperparameters[root_values.name] = root_values
#         else:
#             NotImplementedError()
#
#         for key in sorted(self.hyperparameters):
#             print self.hyperparameters[key]
#         return self.hyperparameters
#
#     def read_object(self, expr, label, condition=None):
#         # Useful pyll objects...
#         if isinstance(expr, configuration_space_module.Hyperparameter):
#             return expr
#         elif expr.name == "hyperopt_param":
#             return self.read_hyperopt_param(expr, condition=condition)
#         elif expr.name == "literal":
#             return self.read_literal(expr, label, condition=condition)
#         elif expr.name == "dict":
#             return self.read_dict(expr, condition=condition)
#         elif expr.name == "pos_args":
#             return self.read_container(expr, condition=condition)
#         elif expr.name == "switch":
#             return self.read_switch(expr, condition=condition)
#         elif expr.name == "float":
#             return self.read_float(expr, condition=condition)
#         elif expr.name == "int":
#             return self.read_int(expr, condition=condition)
#         # Hyperopt distributions
#         elif expr.name == "uniform":
#             return self.read_uniform(expr, label, condition=condition)
#         elif expr.name == "quniform":
#             return self.read_quniform(expr, label, condition=condition)
#         elif expr.name == "loguniform":
#             return self.read_loguniform(expr, label, condition=condition)
#         elif expr.name == "qloguniform":
#             return self.read_qloguniform(expr, label, condition=condition)
#         elif expr.name == "normal":
#             return self.read_normal(expr, label, condition=condition)
#         elif expr.name == "qnormal":
#             return self.read_qnormal(expr, label, condition=condition)
#         elif expr.name == "lognormal":
#             return self.read_lognormal(expr, label, condition=condition)
#         elif expr.name == "qlognormal":
#             return self.read_qlognormal(expr, label, condition=condition)
#         # Mathematical operations
#         elif expr.name in ("floordiv", "pow", "div", "mul", "ceildiv", "max",
#                 "add", "sub"):
#             return self.read_mathematical(expr, condition=condition)
#         elif expr.name == "partial":
#             return self.read_dict(expr, condition=condition)
#         elif expr.name == "getitem":
#             return self.read_getitem(expr, label, condition=condition)
#         elif expr.name == "getattr":
#             return self.read_getattr(expr, condition=condition)
#         elif expr.name == "name":
#             raise NotImplementedError()
#         elif len(expr.named_args) > 0 and len(expr.pos_args) == 0:
#             print "WARNING: encountered unknown pyll node %s, but it has " \
#                   "named arguments which will now be processed" % expr.name
#             return self.read_dict(expr)
#         elif len(expr.named_args) == 0 and len(expr.pos_args) > 0:
#             print "WARNING: encountered unknown pyll node %s, but it has " \
#                   "position arguments which will now be processed" % expr.name
#             return self.read_container(expr)
#         elif len(expr.named_args) > 0 and len(expr.pos_args) > 0:
#             print "WARNING: encountered unknown pyll node %s, but it has " \
#                   "both named and positional arguments which will now be " \
#                   "processed" % expr.name
#             return self.read_container(expr)
#         else:
#             raise Exception("Node name %s not known of %s" % (expr.name, expr))
#
#     def read_literal(self, expr, label, condition=None):
#         # This seems to do no harm...
#         if type(expr._obj) == hyperopt.pyll.base.SymbolTableEntry:
#             pass
#         else:
#             if label is None:
#                 label = str(expr._obj)
#             return configuration_space_module.Constant(label, str(expr._obj),
#                                                        condition)
#
#     def read_float(self, expr, condition=None):
#         # 0 float
#         # 1   hyperopt_param
#         # 2     Literal{b}
#         # 3     uniform
#         # 4       Literal{5}
#         # 5       Literal{15}
#         assert len(expr.inputs()) == 1, expr.inputs()
#         name = expr.inputs()[0].name
#         if name == "literal":
#             raise NotImplementedError()
#         elif name in ("hyperopt_param", "switch"):
#             return self.read_object(expr.inputs()[0], None,
#                                     condition=condition)
#         else:
#             print expr
#             raise NotImplementedError()
#
#     def read_int(self, expr, condition=None):
#         # 0 float
#         # 1   hyperopt_param
#         # 2     Literal{b}
#         # 3     uniform
#         # 4       Literal{5}
#         # 5       Literal{15}
#         assert len(expr.inputs()) == 1, expr.inputs()
#         name = expr.inputs()[0].name
#         if name == "literal":
#             raise NotImplementedError()
#         elif name == "float":
#             param = self.read_object(expr.inputs()[0], None,
#                                     condition=condition)
#             if not isinstance(param, configuration_space_module
#                     .IntegerHyperparameter):
#                 # Rounding ints is done this way in pyll
#                 if isinstance(param, configuration_space_module
#                         .UniformFloatHyperparameter):
#                     param.lower = int(param.lower)
#                     param.upper = int(param.upper)
#                 param = param.to_integer()
#             return param
#         else:
#             print expr
#             raise NotImplementedError()
#
#     def read_hyperopt_param(self, expr, condition=None):
#         # 1   hyperopt_param
#         # 2     Literal{b}
#         # 3     uniform
#         # 4       Literal{5}
#         # 5       Literal{15}
#         assert len(expr.inputs()) == 2, expr.inputs()
#         label = expr.inputs()[0]._obj
#         return self.read_object(expr.inputs()[1], label, condition=condition)
#
#     def read_container(self, expr, condition=None):
#         choices = {}
#         descendants = deque()
#         # descendants.extend(expr.pos_args)
#         descendants.extend(expr.inputs())
#         while len(descendants) > 0:
#             descendant = descendants.popleft()
#             child = self.read_object(descendant, None, condition=condition) #  label not known
#             if child is None:
#                 # this seems to do no harm and can happen e.g. here:
#                 # 0 pos_args
#                 # 1   Literal{<hyperopt.pyll.base.SymbolTableEntry object at 0x7f6afce5b790>}
#                 # 2   switch
#                 continue
#
#             if isinstance(child, dict):
#                 descendants.extend(child.values())
#                 continue
#
#             try:
#                 choices[child.name] = child
#             except AttributeError as e:
#                 print
#                 print expr, child
#                 raise e
#
#         return choices
#
#     def read_dict(self, expr, condition=None):
#         # Only one constant per dict is allowed!
#         values = {}
#         descendants = deque()
#         descendants.extend(expr.named_args)
#         while len(descendants) > 0:
#             descendant = descendants.popleft()
#             child = self.read_object(descendant[1], descendant[0],
#                                      condition=condition)
#             if isinstance(child, dict):
#                 descendants.extend(child.items())
#                 continue
#             try:
#                 values[child.name] = child
#             except AttributeError as e:
#                 for hp_key in self.hyperparameters:
#                     print self.hyperparameters[hp_key]
#                 print "Child", child
#                 print "Expr", expr
#                 raise e
#
#             # The dict must not add values to self.hyperparameters as a
#             # parent switch must be able to control whether this is a real
#             # hyperparameter or just a constant
#
#         # if num_constants != 1:
#         #     raise ValueError("More than one constant value in dict %s" % expr)
#         return values
#
#     def read_switch(self, expr, condition=None):
#         # These checks are really restrictive, they can be relaxed if necessary
#         # e.g. scope.switch(hp.uniform(), []) would be raise an Error but is
#         # perfectly legal
#         # The switch throws away all constants, as the name/value relation of
#         #  them is not unique. Only constants inside a switch, where the key
#         # is the same as the switch name are kept and used as the name of the
#         #  choice
#         assert expr.inputs()[0].name in ("hyperopt_param", "int"), \
#             expr.inputs()[0]
#         randint_node = expr.pos_args[0]
#         if len(randint_node.inputs()) == 1:
#             raise NotImplementedError("You probably try to use a switch "
#                                       "outside a hp.choice, this is not "
#                                       "implemented yet.")
#         if randint_node.inputs()[1].name == "categorical":
#             raise NotImplementedError("This expression seems to be a pchoice:"
#                                       "%s" % expr)
#         assert randint_node.inputs()[1].name == "randint", randint_node.inputs()[1]
#         assert randint_node.inputs()[1].inputs()[0].name == "literal", \
#             randint_node.inputs()[1].inputs()[0]
#         label = randint_node.inputs()[0]._obj
#
#         choices = []
#
#         for i, descendant in enumerate(expr.inputs()[1:]):
#             child = self.read_object(descendant, None) #  label not known
#             choice_name = [str(i)]
#             if type(child) == dict:
#                 # Unfortunately, we have to iterate twice...
#                 for key in child:
#                     if isinstance(child[key], configuration_space_module
#                             .Constant):
#                         if child[key].name == label:
#                             choice_name.append(child[key].value)
#                         else:
#                             print "WARNING: Constants (%s) without a " \
#                                   "hyperparameter will be ignored." % child[key]
#                             self.constants.append(child[key])
#                 if len(choice_name) > 2:
#                     raise ValueError("There must be only one constant per "
#                                      "dictionary for %s and names %s" %
#                                      (expr, choice_name[1:]))
#                 for key in child:
#                     child[key].append_condition(["%s == %s" %
#                                                   (label, choice_name[-1])])
#                     if isinstance(child[key], configuration_space_module
#                             .Constant):
#                         pass
#                     elif self.hyperparameters.has_key(child[key].name):
#                         if child[key] == self.hyperparameters[child[key].name]:
#                             self.hyperparameters[child[key].name]\
#                                 .append_conditions(child[key].conditions)
#                         else:
#                             print "A different hyperparameter with the same " \
#                                   "name has been encountered " \
#                                   "before:\n%s\nnew%s" % \
#                                   (self.hyperparameters[child[key].name], child[key])
#                             raise ValueError()
#                     else:
#                         self.hyperparameters[child[key].name] = child[key]
#
#
#             elif isinstance(child, configuration_space_module.Constant):
#                 child.condition = [["%s == %s" % (label, i)]]
#                 choice_name.append(child.name)
#             elif isinstance(child, configuration_space_module
#                     .NumericalHyperparameter):
#                 child.condition = [["%s == %s" % (label, i)]]
#                 choice_name.append(child.name)
#                 if self.hyperparameters.has_key(child.name):
#                     if child == self.hyperparameters[child.name]:
#                         self.hyperparameters[child.name]\
#                             .append_conditions(child.conditions)
#                     else:
#                         print "A different hyperparameter with the same " \
#                               "name has been encountered " \
#                               "before:\n%s\nnew%s" % \
#                               (self.hyperparameters[child.name], child)
#                         raise ValueError()
#                 else:
#                     self.hyperparameters[child.name] = child
#
#             else:
#                 raise ValueError("This pyll reader only supports dictionaries or"
#                                  " constants "
#                                  "as choices in hp.choice, you provided a %s "
#                                  "for %s in %s" % (type(child), descendant,
#                                                    expr))
#             choices.append(choice_name[-1])
#
#         return configuration_space_module.CategoricalHyperparameter(label, choices)
#
#     def read_getitem(self, expr, label, condition=None):
#         # channels=pyll_getattr(Xcm, 'shape')[1]
#         assert len(expr.pos_args) == 2
#         assert len(expr.named_args) == 0
#         return self.read_container(expr, condition)
#
#     def read_getattr(self, expr, condition=None):
#         assert len(expr.pos_args) == 2
#         assert len(expr.named_args) == 0
#         return self.read_container(expr, condition)
#
#     def read_mathematical(self, expr, condition=None):
#         assert len(expr.pos_args) == 2
#         assert len(expr.named_args) == 0
#         return self.read_container(expr, condition)
#
#     def read_uniform(self, expr, label, condition=None):
#         assert len(expr.inputs()) == 2
#         assert all([input.name == "literal" for input in expr.inputs()]), \
#             expr.inputs()
#         lower = expr.inputs()[0]._obj
#         upper = expr.inputs()[1]._obj
#         return configuration_space_module.UniformFloatHyperparameter(
#             label, lower, upper)
#
#     def read_loguniform(self, expr, label, condition=None):
#         assert len(expr.inputs()) == 2
#         assert all([input.name == "literal" for input in expr.inputs()]), \
#             expr.inputs()
#         lower = np.exp(expr.inputs()[0]._obj)
#         upper = np.exp(expr.inputs()[1]._obj)
#         return configuration_space_module.UniformFloatHyperparameter(
#             label, lower, upper, base=np.e)
#
#     def read_quniform(self, expr, label, condition=None):
#         # Or
#         # 0 quniform
#         # 1   Literal{2.50001}
#         # 2   Literal{8.5}
#         # 3   Literal{1}
#         assert len(expr.inputs()) == 3
#         assert all([input.name == "literal" for input in expr.inputs()]), \
#             expr.inputs()
#         lower = expr.inputs()[0]._obj
#         upper = expr.inputs()[1]._obj
#         q = expr.inputs()[2]._obj
#         if abs(q - 1.0) < 0.0001:
#             # Special case (see top of function)
#             # + avoid numerical issues when rounding param.upper
#             if abs(lower + 0.49999 - np.round(lower, 0)) < 0.0001 and\
#                 abs(upper - 0.5 - np.round(upper - 0.0001, 0)) < 0.0001 and \
#                 abs(q - 1.0) < 0.0001:
#                 lower = np.round(lower, 0)
#                 upper = np.round(upper - 0.0001, 0)
#             else:
#                 print "WARNING: %s has a q-value which suggests the parameter is an Integer," \
#                       "but the ranges could not be properly inferred."
#             return configuration_space_module.UniformIntegerHyperparameter(
#                 label, lower, upper)
#         else:
#             return configuration_space_module.UniformFloatHyperparameter(
#                 label, lower, upper, q=q)
#
#     def read_qloguniform(self, expr, label, condition=None):
#         assert len(expr.inputs()) == 3
#         assert all([input.name == "literal" for input in expr.inputs()]), \
#             expr.inputs()
#         lower = np.exp(expr.inputs()[0]._obj)
#         upper = np.exp(expr.inputs()[1]._obj)
#         q = expr.inputs()[2]._obj
#         if abs(q - 1.0) < 0.0001:
#             # Special case (see top of function)
#             # + avoid numerical issues when rounding param.upper
#             if abs(lower + 0.49999 - np.round(lower, 0)) < 0.0001 and\
#                 abs(upper - 0.5 - np.round(upper - 0.0001, 0)) < 0.0001 and \
#                 abs(q - 1.0) < 0.0001:
#                 lower = np.round(lower, 0)
#                 upper = np.round(upper, 0)
#             else:
#                 print "WARNING: %s has a q-value which suggests the parameter is an Integer," \
#                       "but the ranges could not be properly inferred."
#             return configuration_space_module.UniformIntegerHyperparameter(
#                 label, lower, upper)
#         else:
#             return configuration_space_module.UniformFloatHyperparameter(
#                 label, lower, upper, q=q, base=np.e)
#
#     def read_normal(self, expr, label, condition=None):
#         assert len(expr.inputs()) == 2
#         assert all([input.name == "literal" for input in expr.inputs()]), \
#             expr.inputs()
#         mu = expr.inputs()[0]._obj
#         sigma = expr.inputs()[1]._obj
#         return configuration_space_module.NormalFloatHyperparameter(
#             label, mu, sigma)
#
#     def read_lognormal(self, expr, label, condition=None):
#         # Returns a value drawn according to exp(normal(mu, sigma))
#         assert len(expr.inputs()) == 2
#         assert all([input.name == "literal" for input in expr.inputs()]), \
#             expr.inputs()
#         mu = expr.inputs()[0]._obj
#         sigma = expr.inputs()[1]._obj
#         return configuration_space_module.NormalFloatHyperparameter(
#             label, mu, sigma, base=np.e)
#
#     def read_qnormal(self, expr, label, condition=None):
#         assert len(expr.inputs()) == 3
#         assert all([input.name == "literal" for input in expr.inputs()]), \
#             expr.inputs()
#         mu = expr.inputs()[0]._obj
#         sigma = expr.inputs()[1]._obj
#         q = expr.inputs()[2]._obj
#         if abs(q - 1.0) < 0.0001:
#             return configuration_space_module.NormalIntegerHyperparameter(
#             label, mu, sigma)
#         else:
#             return configuration_space_module.NormalFloatHyperparameter(
#                 label, mu, sigma, q=q)
#
#     def read_qlognormal(self, expr, label, condition=None):
#         # Returns a value like round(exp(normal(mu, sigma)) / q) * q
#         assert len(expr.inputs()) == 3
#         assert all([input.name == "literal" for input in expr.inputs()]), \
#             expr.inputs()
#         mu = expr.inputs()[0]._obj
#         sigma = expr.inputs()[1]._obj
#         q = expr.inputs()[2]._obj
#         if abs(q - 1.0) < 0.0001:
#             return configuration_space_module.NormalIntegerHyperparameter(
#                 label, mu, sigma, base=np.e)
#         else:
#             return configuration_space_module.NormalFloatHyperparameter(
#                 label, mu, sigma, q=q, base=np.e)
#
#
# ################################################################################
# # Write functionality
# def write(configuration_space):
#     pyll_writer = PyllWriter()
#     return pyll_writer.write(configuration_space)
#
#
# class PyllWriter(object):
#     def __init__(self):
#         self.hyperparameters = {}
#
#     def add_hyperparameter(self, hyperparameter):
#         index = "param_%d" % len(self.hyperparameters)
#         self.hyperparameters[hyperparameter.name] = len(self.hyperparameters)
#         return index
#
#     def reset_hyperparameter_countr(self):
#         self.hyperparameters = {}
#
#     def write(self, configuration_space):
#         configuration_space = copy.deepcopy(configuration_space)
#         for key in configuration_space:
#             print configuration_space[key]
#
#         # Name conversions must happen here because the hyperparameter are
#         # later on referenced by this name, the values are converted later!!!
#         for hyperparameter in configuration_space.values():
#             if isinstance(hyperparameter, configuration_space_module
#                     .NumericalHyperparameter) and hyperparameter.base is not \
#                     None:
#                 if abs(hyperparameter.base - np.e) < 0.0000001:
#                     continue
#                 else:
#                     hyperparameter.name = self.convert_name(hyperparameter)
#
#         configuration_dag = configuration_space_module\
#             .create_dag_from_hyperparameters(configuration_space)
#
#         configuration_string = StringIO.StringIO()
#         configuration_string.write('from hyperopt import hp\n')
#         configuration_string.write('import hyperopt.pyll as pyll')
#         configuration_string.write('\n\n')
#
#         strings, hyperparameter_names = self.traverse_dag_depth_first(
#             configuration_dag)
#         for string in strings:
#             configuration_string.write(string)
#             configuration_string.write("\n")
#
#         configuration_string.write('\nspace = {')
#         configuration_string.write(', '
#             .join(['"%s": param_%s' % (name, self.hyperparameters[name])
#                    for name in hyperparameter_names]))
#         configuration_string.write('}\n')
#         configuration_string.seek(0)
#         return configuration_string.getvalue()
#
#     def traverse_dag_depth_first(self, dag):
#         hyperparameter_names = []
#         strings = []
#
#         for name in configuration_space_module.get_dag(dag):
#             hyperparameter = dag.node[name]['hyperparameter']
#             if hyperparameter.conditions == [[]]:
#                 hyperparameter_names.append(name)
#             children = dag[name]
#             _, string = self.write_hyperparameter(hyperparameter, children)
#             strings.append(string)
#
#         return strings, hyperparameter_names
#
#     def write_hyperparameter(self, hyperparameter, children):
#         # Which string generator to call
#         if isinstance(hyperparameter, configuration_space_module.NumericalHyperparameter):
#             generator_name = 'write_'
#             if isinstance(hyperparameter, configuration_space_module.UniformFloatHyperparameter) \
#                          and hyperparameter.q is not None and not (
#                         hyperparameter.q >= hyperparameter.lower or
#                         abs(hyperparameter.q / 2. - hyperparameter.lower) < 0.000001):
#                 raise ValueError()
#
#             if hyperparameter.base is None and hyperparameter.q is None:
#                 pass
#             elif hyperparameter.base is None and hyperparameter.q is not None:
#                 generator_name += 'q'
#             elif abs(hyperparameter.base - math.e) < 0.000001:
#                 if hyperparameter.q is not None:
#                     generator_name += 'q'
#                 generator_name += 'log'
#             else:
#                 pass
#
#             if children is not None and len(children) > 0:
#                 raise NotImplementedError()
#
#             if type(hyperparameter) in (configuration_space_module
#                                                 .NormalFloatHyperparameter,
#                                         configuration_space_module
#                                                 .NormalIntegerHyperparameter):
#                 generator_name += 'normal'
#             else:
#                 generator_name += 'uniform'
#
#             if hyperparameter.base is None or abs(hyperparameter.base - \
#                     math.e) < 0.000001:
#                 if isinstance(hyperparameter, \
#                     configuration_space_module.IntegerHyperparameter):
#                     generator_name += "_int"
#
#             name, string = getattr(self, generator_name)(hyperparameter)
#
#         elif isinstance(hyperparameter, configuration_space_module.CategoricalHyperparameter):
#                 name, string = self.write_choice(hyperparameter, children)
#
#         else:
#             raise NotImplementedError()
#
#         return name, string
#
#     def write_choice(self, parameter, children):
#         name = parameter.name
#         choices = dict()
#         for choice in parameter.choices:
#             choices[choice] = dict()
#
#         for key in children:
#             child = children[key]
#             operator = child['condition'][1]
#             if operator == "==":
#                 value = child['condition'][2]
#                 choices[value][key] = child
#             elif operator == "in":
#                 values = child['condition'][2].replace("{", "").replace("}", "")
#                 values = values.split(",")
#                 for value in values:
#                     choices[value][key] = child
#             else:
#                 raise NotImplementedError("You used an operator to specifiy "
#                                           "conditions which is not yet "
#                                           "implemented: %s" % operator)
#
#         index = "param_%d" % len(self.hyperparameters)
#         self.hyperparameters[name] = len(self.hyperparameters)
#         return_string = '%s = hp.choice("%s", [\n' % (index, name)
#         for choice in sorted(choices):
#             return_string += '    {'
#             return_string += '"%s": "%s", ' % (name, choice)
#             for key in sorted(choices[choice]):
#                 return_string += '"%s": param_%s, ' % \
#                                  (key, self.hyperparameters[key])
#             return_string += '},\n'
#         return_string += '    ])'
#
#         return name, return_string
#
#     def write_uniform(self, parameter):
#         name = parameter.name
#         # Special treatment of hyperparameters which potentially have a LOG
#         # and a Q in their name...
#         if parameter.q is not None or parameter.base is not None or \
#                 isinstance(parameter, configuration_space_module.IntegerHyperparameter):
#             if parameter.q is not None and abs(parameter.q - 1.0) < 0.00000001 or \
#                     isinstance(parameter, configuration_space_module.IntegerHyperparameter):
#                 parameter.lower = float(parameter.lower) - 0.49999
#                 parameter.upper = float(parameter.upper) + 0.5
#             elif parameter.q is not None:
#                 parameter.lower = float(parameter.lower) - (parameter.q / 2.0 - 0.0001)
#                 parameter.upper = float(parameter.upper) + (parameter.q / 2.0)
#
#         lower, upper = self.get_bounds_as_exponent(parameter)
#         index = self.add_hyperparameter(parameter)
#         return name, '%s = hp.uniform("%s", %s, %s)' % \
#             (index, name, lower, upper)
#
#     def write_uniform_int(self, parameter):
#         name = parameter.name
#         # As like for all other uniform_int parameters, subtracting a number
#         # is fine if the parameter is optimized on log scale as
#         # a) it either is zero and thus illegal before subtraction
#         # b) it is > 1 and thus still legal after subtraction
#         parameter.lower = float(parameter.lower) - 0.49999
#         parameter.upper = float(parameter.upper) + 0.5
#         lower, upper = self.get_bounds_as_exponent(parameter)
#         q = 1.0
#         index = self.add_hyperparameter(parameter)
#         return name, '%s = pyll.scope.int(hp.quniform("%s", %s, %s, %s))' \
#             % (index, name, lower, upper, q)
#
#     def write_quniform(self, parameter):
#         name = parameter.name
#         parameter.lower = float(parameter.lower) - (parameter.q / 2.0 - 0.0001)
#         parameter.upper = float(parameter.upper) + (parameter.q / 2.0)
#         lower, upper = self.get_bounds_as_exponent(parameter)
#         q = float(parameter.q)
#         index = self.add_hyperparameter(parameter)
#         return name, '%s = hp.quniform("%s", %s, %s, %s)' % \
#             (index, name, lower, upper, q)
#
#     def write_quniform_int(self, parameter):
#         name = parameter.name
#         parameter.lower = float(parameter.lower) - 0.49999
#         parameter.upper = float(parameter.upper) + 0.5
#         lower, upper = self.get_bounds_as_exponent(parameter)
#         q = float(parameter.q) if parameter.q is not None else 1.0
#         assert abs(q - int(q)) < 0.000001
#         index = self.add_hyperparameter(parameter)
#         return name, '%s = pyll.scope.int(hp.quniform("%s", %s, %s, %s))' \
#             % (index, name, lower, upper, q)
#
#     def write_loguniform(self, parameter):
#         name = parameter.name
#         lower, upper = self.get_bounds_as_exponent(parameter)
#         index = self.add_hyperparameter(parameter)
#         return name, '%s = hp.loguniform("%s", %s, %s)' % \
#             (index, name, lower, upper)
#
#     def write_loguniform_int(self, parameter):
#         name = parameter.name
#         parameter.lower = float(parameter.lower) - 0.49999
#         parameter.upper = float(parameter.upper) + 0.5
#         lower, upper = self.get_bounds_as_exponent(parameter)
#         q = 1.0
#         index = self.add_hyperparameter(parameter)
#         return name, '%s = pyll.scope.int(hp.qloguniform("%s", %s, %s, %s))' % \
#             (index, name, lower, upper, q)
#
#     def write_qloguniform(self, parameter):
#         name = parameter.name
#         parameter.lower = float(parameter.lower) - (parameter.q / 2.0 - 0.0001)
#         parameter.upper = float(parameter.upper) + (parameter.q / 2.0)
#         lower, upper = self.get_bounds_as_exponent(parameter)
#         q = float(parameter.q)
#         index = self.add_hyperparameter(parameter)
#         return name, '%s = hp.qloguniform("%s", %s, %s, %s)' % \
#             (index, name, lower, upper, q)
#
#     def write_qloguniform_int(self, parameter):
#         name = parameter.name
#         parameter.lower = float(parameter.lower) - 0.49999
#         parameter.upper = float(parameter.upper) + 0.5
#         lower, upper = self.get_bounds_as_exponent(parameter)
#         q = float(parameter.q)
#         assert abs(q - int(q)) < 0.000001
#         index = self.add_hyperparameter(parameter)
#         return name, '%s = pyll.scope.int(hp.qloguniform("%s", %s, %s, %s))' % \
#             (index, name, lower, upper, q)
#
#     def write_normal(self, parameter):
#         name = parameter.name
#         mu = float(parameter.mu)
#         sigma = float(parameter.sigma)
#         index = self.add_hyperparameter(parameter)
#         return name, '%s = hp.normal("%s", %s, %s)' % (index, name, mu, sigma)
#
#     def write_normal_int(self, parameter):
#         name = parameter.name
#         mu = float(parameter.mu)
#         sigma = float(parameter.sigma)
#         index = self.add_hyperparameter(parameter)
#         return name, '%s = pyll.scope.int(hp.qnormal("%s", %s, %s, 1.0))' % \
#                      (index, name, mu, sigma)
#
#     def write_qnormal(self, parameter):
#         name = parameter.name
#         mu = float(parameter.mu)
#         sigma = float(parameter.sigma)
#         q = float(parameter.q)
#         index = self.add_hyperparameter(parameter)
#         return name, '%s = hp.qnormal("%s", %s, %s, %s)' % (index, name, mu, sigma, q)
#
#     def write_qnormal_int(self, parameter):
#         name = parameter.name
#         mu = float(parameter.mu)
#         sigma = float(parameter.sigma)
#         q = float(parameter.q)
#         assert abs(q - int(q)) < 0.000001
#         index = self.add_hyperparameter(parameter)
#         return name, '%s = pyll.scope.int(hp.qnormal("%s", %s, %s, %s))' % \
#                      (index, name, mu, sigma, q)
#
#     def write_lognormal(self, parameter):
#         name = parameter.name
#         mu = float(parameter.mu)
#         sigma = float(parameter.sigma)
#         index = self.add_hyperparameter(parameter)
#         return name, '%s = hp.lognormal("%s", %s, %s)' % (index, name, mu, sigma)
#
#     def write_lognormal_int(self, parameter):
#         name = parameter.name
#         mu = float(parameter.mu)
#         sigma = float(parameter.sigma)
#         index = self.add_hyperparameter(parameter)
#         return name, '%s = pyll.scope.int(hp.qlognormal("%s", %s, %s, 1.0))' \
#                      % (index, name, mu, sigma)
#
#     def write_qlognormal(self, parameter):
#         name = parameter.name
#         mu = float(parameter.mu)
#         sigma = float(parameter.sigma)
#         q = float(parameter.q)
#         index = self.add_hyperparameter(parameter)
#         return name, '%s = hp.qlognormal("%s", %s, %s, %s)' % \
#                      (index, name, mu, sigma, q)
#
#     def write_qlognormal_int(self, parameter):
#         name = parameter.name
#         mu = float(parameter.mu)
#         sigma = float(parameter.sigma)
#         q = float(parameter.q)
#         assert abs(q - int(q)) < 0.000001
#         index = self.add_hyperparameter(parameter)
#         return name, '%s = pyll.scope.int(hp.qlognormal("%s", %s, %s, %s))' % \
#                      (index, name, mu, sigma, q)
#
#     def convert_name(self, parameter):
#         """Add the LOG_ instruction to the name."""
#         base = parameter.base
#         q = parameter.q
#         name = parameter.name
#         if base is not None:
#             if q is not None:
#                 name = "Q%f_%s" % (q, name)
#             elif isinstance(parameter, configuration_space_module
#                                        .IntegerHyperparameter):
#                 name = "Q1_%s" % name
#             if abs(base - np.e) < 0.000001:
#                 return "LOG_%s" % name
#             elif abs(base - 10.0) < 0.000001:
#                 return "LOG10_%s" % name
#             elif abs(base - 2.0) < 0.000001:
#                 return "LOG2_%s" % name
#             elif abs(base - int(base)) < 0.000001:
#                 return "LOG%d_%s" % (int(base), name)
#             else:
#                 raise ValueError("Base Value %s not allowed" % str(base))
#         return name
#
#     def get_bounds_as_exponent(self, parameter):
#         """Return the exponent of lower and upper bound of a parameter.
#
#         Return the exponent of lower and upper bound of a parameter. If the
#         base of the parameter is None, return the lower and upper bound."""
#         base = parameter.base
#         lower = parameter.lower
#         upper = parameter.upper
#
#         if base is not None:
#             if parameter.lower <= 0:
#                 raise ValueError("Lower value for a log scale parameter is not "
#                              "allowed to be <= 0 (%f)" % parameter.lower)
#
#             if abs(base - np.e) < 0.000001:
#                 return np.log(lower), np.log(upper)
#             elif abs(base - 10.0) < 0.000001:
#                 return np.log10(lower), np.log10(upper)
#             elif abs(base - 2.0) < 0.000001:
#                 return np.log2(lower), np.log2(upper)
#             elif abs(base - int(base)) < 0.000001:
#                 lower = np.log10(lower) / np.log10(base)
#                 upper = np.log10(upper) / np.log10(base)
#                 return lower, upper
#             else:
#                 raise ValueError("Base Value %s not allowed" % str(base))
#
#         return lower, upper


# Copyright (c) 2014-2017, ConfigSpace developers
# Matthias Feurer
# Katharina Eggensperger
# Mohsin Ali
# and others (see commit history).
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest

from ConfigSpace.configuration_space import ConfigurationSpace
import ConfigSpace.read_and_write.irace as irace

from ConfigSpace.hyperparameters import CategoricalHyperparameter, \
    UniformIntegerHyperparameter, UniformFloatHyperparameter, \
    OrdinalHyperparameter
from ConfigSpace.conditions import EqualsCondition, InCondition, \
    AndConjunction, OrConjunction
from ConfigSpace.forbidden import ForbiddenInClause, ForbiddenEqualsClause, \
    ForbiddenAndConjunction


# Copyright (c) 2014-2016, ConfigSpace developers
# Matthias Feurer
# Katharina Eggensperger
# and others (see commit history).
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


int_a = UniformIntegerHyperparameter("int_a", -1, 6)

class TestIraceWriter(unittest.TestCase):
    '''
    Test IRACE writer
    '''

    def test_write_illegal_argument(self):
        sp = {"a": int_a}
        self.assertRaisesRegexp(TypeError, "irace.write expects an "
                                "instance of "
                                "<class "
                                "'ConfigSpace.configuration_"
                                "space.ConfigurationSpace'>, you provided "
                                "'<(type|class) 'dict'>'", irace.write, sp)

    def test_write_int(self):
        expected = "int_a '--int_a ' i (-1, 6)\n"
        cs = ConfigurationSpace()
        cs.add_hyperparameter(int_a)
        value = irace.write(cs)
        self.assertEqual(expected, value)



    def test_write_float(self):
        expected = "float_a '--float_a ' r (16.000000, 1024.000000)\n"
        cs = ConfigurationSpace()
        cs.add_hyperparameter(
            UniformFloatHyperparameter("float_a", 16, 1024))
        value = irace.write(cs)
        self.assertEqual(expected, value)


    def test_write_categorical(self):
        expected = "cat_a '--cat_a ' c {a,b,c}\n"
        cs = ConfigurationSpace()
        cs.add_hyperparameter(
            CategoricalHyperparameter("cat_a", ["a", "b", "c"]))
        value = irace.write(cs)
        self.assertEqual(expected, value)


    def test_write_ordinal(self):
        expected = "ord_a '--ord_a ' o {a,b,3}\n"
        cs = ConfigurationSpace()
        cs.add_hyperparameter(
            OrdinalHyperparameter("ord_a", ["a", "b", 3]))
        value = irace.write(cs)
        self.assertEqual(expected, value)

    def test_write_equals_condition_categorical(self):
        expected = "ls '--ls ' c {sa,ca,ny}\ntemp '--temp ' r (0.500000, 1.000000)|  ls==sa\n"

        temp = UniformFloatHyperparameter("temp", 0.5, 1)
        ls = CategoricalHyperparameter("ls", ["sa", "ca", "ny"], "sa")

        cs = ConfigurationSpace()
        cs.add_hyperparameter(temp)
        cs.add_hyperparameter(ls)
        c1 = EqualsCondition(temp, ls, 'sa')
        cs.add_condition(c1)
        value = irace.write(cs)
        self.assertEqual(expected, value)

    def test_write_equals_condition_numerical(self):
        expected = "temp '--temp ' i (1, 2)\nls '--ls ' c {sa,ca,ny}|  temp==2\n"

        temp = UniformIntegerHyperparameter("temp", 1, 2)
        ls = CategoricalHyperparameter("ls", ["sa", "ca", "ny"], "sa")

        cs = ConfigurationSpace()
        cs.add_hyperparameter(temp)
        cs.add_hyperparameter(ls)
        c1 = EqualsCondition(ls, temp, 2)
        cs.add_condition(c1)
        value = irace.write(cs)
        self.assertEqual(expected, value)

    def test_write_in_condition(self):
        expected = "ls '--ls ' c {sa,ca,ny}\ntemp '--temp ' r (0.500000, 1.000000)|  ls  %in%  c(sa,ca)\n"

        temp = UniformFloatHyperparameter("temp", 0.5, 1)
        ls = CategoricalHyperparameter("ls", ["sa", "ca", "ny"], "sa")

        cs = ConfigurationSpace()
        cs.add_hyperparameter(temp)
        cs.add_hyperparameter(ls)
        c1 = InCondition(temp, ls, ['sa','ca'])
        cs.add_condition(c1)
        value = irace.write(cs)
        self.assertEqual(expected, value)


    def test_write_AndConjunction_condition(self):
        expected = "lp '--lp ' c {mi,bo}\nls '--ls ' c {sa,ca,ny}\ntemp '--temp ' r (0.500000, 1.000000)|  ls  %in%  c(sa,ca)  &&  lp  %in%  c(bo)\n"

        temp = UniformFloatHyperparameter("temp", 0.5, 1)
        ls = CategoricalHyperparameter("ls", ["sa", "ca", "ny"], "sa")
        lp = CategoricalHyperparameter("lp", ["mi", "bo"], "bo")

        cs = ConfigurationSpace()
        cs.add_hyperparameter(temp)
        cs.add_hyperparameter(lp)
        cs.add_hyperparameter(ls)

        c1 = InCondition(temp, ls, ['sa','ca'])
        c2 = InCondition(temp, lp, ['bo'])
        c3 = AndConjunction(c1, c2)
        cs.add_condition(c3)
        value = irace.write(cs)
        self.assertEqual(expected, value)

    def test_write_OrConjunction_condition(self):
        import numpy as np
        expected = "lp '--lp ' c {mi,bo}\ntemp '--temp ' r (2.000000, 5.000000)\nls '--ls ' c {sa,ca,ny}|  temp==3.0  ||  lp  %in%  c(bo)\n"

        temp = UniformFloatHyperparameter("temp", np.exp(2), np.exp(5), log=True)
        ls = CategoricalHyperparameter("ls", ["sa", "ca", "ny"], "sa")
        lp = CategoricalHyperparameter("lp", ["mi", "bo"], "bo")

        cs = ConfigurationSpace()
        cs.add_hyperparameter(temp)
        cs.add_hyperparameter(lp)
        cs.add_hyperparameter(ls)

        c1 = EqualsCondition(ls, temp, np.exp(3))
        c2 = InCondition(ls, lp, ['bo'])
        c3 = OrConjunction(c1, c2)
        cs.add_condition(c3)
        value = irace.write(cs)
        self.assertEqual(expected, value)

    def test_write_forbidden(self):
        cs =ConfigurationSpace()

        hp1 = CategoricalHyperparameter("parent", [0, 1])
        hp2 = UniformIntegerHyperparameter("child", 0, 2)
        hp3 = UniformIntegerHyperparameter("child2", 0, 2)
        hp4 = UniformIntegerHyperparameter("child3", 0, 2)
        hp5 = CategoricalHyperparameter("child4", [4,5,6,7])

        cs.add_hyperparameters([hp1, hp2, hp3, hp4, hp5])

        forb2 = ForbiddenEqualsClause(hp1, 1)
        forb3 = ForbiddenInClause(hp2, range(2, 3))
        forb4 = ForbiddenInClause(hp3, range(2, 3))
        forb5 = ForbiddenInClause(hp4, range(2, 3))
        forb6 = ForbiddenInClause(hp5, [6,7])

        and1 = ForbiddenAndConjunction(forb2, forb3)
        and2 = ForbiddenAndConjunction(forb2, forb4)
        and3 = ForbiddenAndConjunction(forb2, forb5)

        cs.add_forbidden_clauses([forb2, forb3, forb4, forb5, forb6, and1, and2, and3])

        value = irace.write(cs) # generates file called forbidden.txt


    def test_write_log_int(self):
        expected = "int_log '--int_log ' i (2, 4)\n"
        int_log = UniformIntegerHyperparameter("int_log", 10, 100, log=True)
        cs = ConfigurationSpace()
        cs.add_hyperparameter(int_log)
        value = irace.write(cs)
        self.assertEqual(expected, value)

    def test_write_log_float(self):
        import numpy as np
        expected = "float_log '--float_log ' r (2.000000, 5.000000)\n"
        float_log = UniformFloatHyperparameter("float_log", np.exp(2), np.exp(5), log=True)
        cs = ConfigurationSpace()
        cs.add_hyperparameter(float_log)
        value = irace.write(cs)
        self.assertEqual(expected, value)



# Copyright (c) 2014-2016, ConfigSpace developers
# Matthias Feurer
# Katharina Eggensperger
# and others (see commit history).
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from abc import ABCMeta, abstractmethod
import copy
from itertools import combinations
from typing import Any, List, Union, Tuple, Dict, Optional
import operator

from libc.stdlib cimport malloc, free

import numpy as np
cimport numpy as np

# We now need to fix a datatype for our arrays. I've used the variable
# DTYPE for this, which is assigned to the usual NumPy runtime
# type info object.
#DTYPE = np.float
# "ctypedef" assigns a corresponding compile-time type to DTYPE_t. For
# every type in the numpy module there's a corresponding compile-time
# type with a _t-suffix.
#ctypedef np.float_t DTYPE_t

import io
from functools import reduce
from ConfigSpace.hyperparameters import NumericalHyperparameter, OrdinalHyperparameter
from ConfigSpace.hyperparameters cimport Hyperparameter


cdef class ConditionComponent(object):

    def __init__(self) -> None:
        pass

    def __repr__(self) -> str:
        pass

    def set_vector_idx(self, hyperparameter_to_idx) -> None:
        pass

    def get_children_vector(self) -> List[int]:
        pass

    def get_parents_vector(self) -> List[int]:
        pass

    def get_children(self) -> List['ConditionComponent']:
        pass

    def get_parents(self) -> List['ConditionComponent']:
        pass

    def get_descendant_literal_conditions(self) ->List['AbstractCondition']:
        pass

    def evaluate(self, instantiated_parent_hyperparameter: Dict[str, Union[None, int, float, str]]) -> bool:
        pass

    def evaluate_vector(self, instantiated_vector):
        return bool(self._evaluate_vector(instantiated_vector))

    cdef int _evaluate_vector(self, np.ndarray value):
        pass

    def __hash__(self) -> int:
        """Override the default hash behavior (that returns the id or the object)"""
        return hash(tuple(sorted(self.__dict__.items())))


cdef class AbstractCondition(ConditionComponent):
    cdef public Hyperparameter child
    cdef public Hyperparameter parent
    cdef public int child_vector_id
    cdef public int parent_vector_id
    cdef public value
    cdef public DTYPE_t vector_value

    def __init__(self, child: Hyperparameter, parent: Hyperparameter) -> None:
        if not isinstance(child, Hyperparameter):
            raise ValueError("Argument 'child' is not an instance of "
                             "HPOlibConfigSpace.hyperparameter.Hyperparameter.")
        if not isinstance(parent, Hyperparameter):
            raise ValueError("Argument 'parent' is not an instance of "
                             "HPOlibConfigSpace.hyperparameter.Hyperparameter.")
        if child == parent:
            raise ValueError("The child and parent hyperparameter must be "
                             "different hyperparameters.")
        self.child = child
        self.parent = parent
        self.child_vector_id = -1
        self.parent_vector_id = -1

    def __richcmp__(self, other: Any, int op):
        """Override the default Equals behavior
        There are no separate methods for the individual rich comparison operations (__eq__(), __le__(), etc.).
         Instead there is a single method __richcmp__() which takes an integer indicating which operation is to be performed, as follows:
        < 	0
        == 2
        > 	4
        <=	1
        !=	3
        >=	5
        """
        if isinstance(other, self.__class__):

            if op == 2:
                if self.child != other.child:
                    return False
                elif self.parent != other.parent:
                    return False
                return self.value == other.value

            elif op == 3:
                if self.child != other.child or self.parent != other.parent or self.value != other.value:
                    return True
                else:
                    return False

        return NotImplemented

    def set_vector_idx(self, hyperparameter_to_idx: dict):
        self.child_vector_id = hyperparameter_to_idx[self.child.name]
        self.parent_vector_id = hyperparameter_to_idx[self.parent.name]

    def get_children_vector(self) -> List[int]:
        return [self.child_vector_id]

    def get_parents_vector(self) -> List[int]:
        return [self.parent_vector_id]

    def get_children(self) -> List[Hyperparameter]:
        return [self.child]

    def get_parents(self) -> List[Hyperparameter]:
        return [self.parent]

    def get_descendant_literal_conditions(self) -> List['AbstractCondition']:
        return [self]

    def evaluate(self, instantiated_parent_hyperparameter: Dict[str, Union[int, float, str]]) -> bool:
        hp_name = self.parent.name
        return self._evaluate(instantiated_parent_hyperparameter[hp_name])

    cdef int _evaluate_vector(self, np.ndarray instantiated_vector):
        if self.parent_vector_id is None:
            raise ValueError("Parent vector id should not be None when calling evaluate vector")
        return self._inner_evaluate_vector(instantiated_vector[self.parent_vector_id])

    def _evaluate(self, instantiated_parent_hyperparameter: Union[str, int, float]) -> bool:
        pass

    cdef int _inner_evaluate_vector(self, DTYPE_t value):
        pass


cdef class EqualsCondition(AbstractCondition):

    def __init__(self, child: Hyperparameter, parent: Hyperparameter, value: Union[str, float, int]) -> None:
        super(EqualsCondition, self).__init__(child, parent)
        if not parent.is_legal(value):
            raise ValueError("Hyperparameter '%s' is "
                             "conditional on the illegal value '%s' of "
                             "its parent hyperparameter '%s'" %
                             (child.name, value, parent.name))
        self.value = value
        self.vector_value = self.parent._inverse_transform(self.value)

    def __repr__(self) -> str:
        return "%s | %s == %s" % (self.child.name, self.parent.name,
                                  repr(self.value))

    def __copy__(self):
            return self.__class__(
                child=copy.copy(self.child),
                parent=copy.copy(self.parent),
                value=copy.copy(self.value),
            )

    def _evaluate(self, value: Union[str, float, int]) -> bool:
        # No need to check if the value to compare is a legal value; either it
        # is equal (and thus legal), or it would evaluate to False anyway

        cmp = self.parent.compare(value, self.value)
        if cmp == 0:
            return True
        else:
            return False

    cdef int _inner_evaluate_vector(self, DTYPE_t value):
        # No need to check if the value to compare is a legal value; either it
        # is equal (and thus legal), or it would evaluate to False anyway

        cdef int cmp = self.parent.compare_vector(value, self.vector_value)
        if cmp == 0:
            return True
        else:
            return False


cdef class NotEqualsCondition(AbstractCondition):
    def __init__(self, child: Hyperparameter, parent: Hyperparameter, value: Union[str, float, int]) -> None:
        super(NotEqualsCondition, self).__init__(child, parent)
        if not parent.is_legal(value):
            raise ValueError("Hyperparameter '%s' is "
                             "conditional on the illegal value '%s' of "
                             "its parent hyperparameter '%s'" %
                             (child.name, value, parent.name))
        self.value = value
        self.vector_value = self.parent._inverse_transform(self.value)

    def __repr__(self) -> str:
        return "%s | %s != %s" % (self.child.name, self.parent.name,
                                  repr(self.value))

    def __copy__(self):
            return self.__class__(
                child=copy.copy(self.child),
                parent=copy.copy(self.parent),
                value=copy.copy(self.value),
            )

    def _evaluate(self, value: Union[str, float, int]) -> bool:
        if not self.parent.is_legal(value):
            return False

        cmp = self.parent.compare(value, self.value)
        if cmp != 0:
            return True
        else:
            return False

    cdef int _inner_evaluate_vector(self, DTYPE_t value):
        if not self.parent.is_legal_vector(value):
            return False

        cdef int cmp = self.parent.compare_vector(value, self.vector_value)
        if cmp != 0:
            return True
        else:
            return False


cdef class LessThanCondition(AbstractCondition):
    def __init__(self, child: Hyperparameter, parent: Hyperparameter, value: Union[str, float, int]) -> None:
        super(LessThanCondition, self).__init__(child, parent)
        self.parent.allow_greater_less_comparison()
        if not parent.is_legal(value):
            raise ValueError("Hyperparameter '%s' is "
                             "conditional on the illegal value '%s' of "
                             "its parent hyperparameter '%s'" %
                             (child.name, value, parent.name))
        self.value = value
        self.vector_value = self.parent._inverse_transform(self.value)

    def __repr__(self) -> str:
        return "%s | %s < %s" % (self.child.name, self.parent.name,
                                 repr(self.value))

    def __copy__(self):
            return self.__class__(
                child=copy.copy(self.child),
                parent=copy.copy(self.parent),
                value=copy.copy(self.value),
            )

    def _evaluate(self, value: Union[str, float, int]) -> bool:
        if not self.parent.is_legal(value):
            return False

        cmp = self.parent.compare(value, self.value)
        if cmp == -1:
            return True
        else:
            return False

    cdef int _inner_evaluate_vector(self, DTYPE_t value):
        if not self.parent.is_legal_vector(value):
            return False

        cdef int cmp = self.parent.compare_vector(value, self.vector_value)
        if cmp == -1:
            return True
        else:
            return False


cdef class GreaterThanCondition(AbstractCondition):
    def __init__(self, child: Hyperparameter, parent: Hyperparameter, value: Union[str, float, int]) -> None:
        super(GreaterThanCondition, self).__init__(child, parent)
        self.parent.allow_greater_less_comparison()
        if not parent.is_legal(value):
            raise ValueError("Hyperparameter '%s' is "
                             "conditional on the illegal value '%s' of "
                             "its parent hyperparameter '%s'" %
                             (child.name, value, parent.name))
        self.value = value
        self.vector_value = self.parent._inverse_transform(self.value)

    def __repr__(self) -> str:
        return "%s | %s > %s" % (self.child.name, self.parent.name,
                                 repr(self.value))

    def __copy__(self):
            return self.__class__(
                child=copy.copy(self.child),
                parent=copy.copy(self.parent),
                value=copy.copy(self.value),
            )

    def _evaluate(self, value: Union[str, float, int]) -> bool:
        if not self.parent.is_legal(value):
            return False

        cmp = self.parent.compare(value, self.value)
        if cmp == 1:
            return True
        else:
            return False

    cdef int _inner_evaluate_vector(self, DTYPE_t value):
        if not self.parent.is_legal_vector(value):
            return False

        cdef int cmp = self.parent.compare_vector(value, self.vector_value)
        if cmp == 1:
            return True
        else:
            return False

cdef class InCondition(AbstractCondition):
    cdef public values
    cdef public vector_values

    def __init__(self, child: Hyperparameter, parent: Hyperparameter, values: List[Union[str, float, int]]) -> None:
        super(InCondition, self).__init__(child, parent)
        for value in values:
            if not parent.is_legal(value):
                raise ValueError("Hyperparameter '%s' is "
                                 "conditional on the illegal value '%s' of "
                                 "its parent hyperparameter '%s'" %
                                 (child.name, value, parent.name))
        self.values = values
        self.value = values
        self.vector_values = [self.parent._inverse_transform(value) for value in self.values]

    def __richcmp__(self, other: Any, int op):
        """Override the default Equals behavior
        There are no separate methods for the individual rich comparison operations (__eq__(), __le__(), etc.).
         Instead there is a single method __richcmp__() which takes an integer indicating which operation is to be performed, as follows:
        < 	0
        == 2
        > 	4
        <=	1
        !=	3
        >=	5
        """
        if isinstance(other, self.__class__):


            if op == 2:
                if self.child != other.child:
                    return False
                elif self.parent != other.parent:
                    return False
                return self.values == other.values

            elif op == 3:
                if self.child != other.child or self.parent != other.parent or self.values != other.values:
                    return True
                else:
                    return False

        return NotImplemented

    def __repr__(self) -> str:
        return "%s | %s in {%s}" % (self.child.name, self.parent.name,
                                    ", ".join(
                                        [repr(value) for value in self.values]))

    def _evaluate(self, value: Union[str, float, int]) -> bool:
        return value in self.values

    cdef int _inner_evaluate_vector(self, DTYPE_t value):
        return value in self.vector_values


cdef class AbstractConjunction(ConditionComponent):
    cdef public tuple components
    cdef int n_components
    cdef tuple dlcs

    def __init__(self, *args: AbstractCondition) -> None:
        super(AbstractConjunction, self).__init__()
        self.components = args
        self.n_components = len(self.components)
        self.dlcs = self.get_descendant_literal_conditions()

        # Test the classes
        for idx, component in enumerate(self.components):
            if not isinstance(component, ConditionComponent):
                raise TypeError("Argument #%d is not an instance of %s, "
                                "but %s" % (
                                    idx, ConditionComponent, type(component)))

        # Test that all conjunctions and conditions have the same child!
        children = self.get_children()
        for c1, c2 in combinations(children, 2):
            if c1 != c2:
                raise ValueError("All Conjunctions and Conditions must have "
                                 "the same child.")

    def __richcmp__(self, other: Any, int op):
        """Override the default Equals behavior
        There are no separate methods for the individual rich comparison operations (__eq__(), __le__(), etc.).
         Instead there is a single method __richcmp__() which takes an integer indicating which operation is to be performed, as follows:
        < 	0
        == 2
        > 	4
        <=	1
        !=	3
        >=	5
        """
        if isinstance(other, self.__class__):
            if len(self.components) != len(other.components):
                if op == 2:
                    return False
                if op == 3:
                    return True
                else:
                    return NotImplemented

            for component, other_component in \
                    zip(self.components, other.components):
                eq = component == other_component
                if op == 2:
                    if not eq:
                        return False
                elif op == 3:
                    if eq:
                        return False
                else:
                    raise NotImplemented
            return True

        return NotImplemented

    def __copy__(self):
        return self.__class__(*[copy.copy(comp) for comp in self.components])

    def get_descendant_literal_conditions(self) -> Tuple[AbstractCondition]:
        children = []  # type: List[AbstractCondition]
        for component in self.components:
            if isinstance(component, AbstractConjunction):
                children.extend(component.get_descendant_literal_conditions())
            else:
                children.append(component)
        return tuple(children)

    def set_vector_idx(self, hyperparameter_to_idx: dict):
        for component in self.components:
            component.set_vector_idx(hyperparameter_to_idx)

    def get_children_vector(self) -> List[int]:
        children_vector = []
        for component in self.components:
            children_vector.extend(component.get_children_vector())
        return children_vector

    def get_parents_vector(self) -> List[int]:
        parents_vector = []
        for component in self.components:
            parents_vector.extend(component.get_parents_vector())
        return parents_vector

    def get_children(self) -> List[ConditionComponent]:
        children = []  # type: List[ConditionComponent]
        for component in self.components:
            children.extend(component.get_children())
        return children

    def get_parents(self) -> List[ConditionComponent]:
        parents = []  # type: List[ConditionComponent]
        for component in self.components:
            parents.extend(component.get_parents())
        return parents

    def evaluate(self, instantiated_hyperparameters: Dict[str, Union[None, int, float, str]]) -> bool:
        cdef int* arrptr
        arrptr = <int*> malloc(sizeof(int) * self.n_components)

        # Then, check if all parents were passed
        conditions = self.dlcs
        for condition in conditions:
            if condition.parent.name not in instantiated_hyperparameters:
                raise ValueError("Evaluate must be called with all "
                                 "instanstatiated parent hyperparameters in "
                                 "the conjunction; you are (at least) missing "
                                 "'%s'" % condition.parent.name)

        # Finally, call evaluate for all direct descendents and combine the
        # outcomes
        for i, component in enumerate(self.components):
            e = component.evaluate(instantiated_hyperparameters)
            arrptr[i] = (e)

        rval = self._evaluate(self.n_components, arrptr)
        free(arrptr)
        return rval

    cdef int _evaluate_vector(self, np.ndarray instantiated_vector):
        cdef ConditionComponent component
        cdef int e
        cdef int rval
        cdef int* arrptr
        arrptr = <int*> malloc(sizeof(int) * self.n_components)

        # Finally, call evaluate for all direct descendents and combine the
        # outcomes
        for i in range(self.n_components):
            component = self.components[i]
            e = component._evaluate_vector(instantiated_vector)
            arrptr[i] = e

        rval = self._evaluate(self.n_components, arrptr)
        free(arrptr)
        return rval

    cdef int _evaluate(self, int I, int* evaluations):
        pass


cdef class AndConjunction(AbstractConjunction):
    # TODO: test if an AndConjunction results in an illegal state or a
    # Tautology! -> SAT solver
    def __init__(self, *args: AbstractCondition) -> None:
        if len(args) < 2:
            raise ValueError("AndConjunction must at least have two "
                             "Conditions.")
        super(AndConjunction, self).__init__(*args)

    def __repr__(self) -> str:
        retval = io.StringIO()
        retval.write("(")
        for idx, component in enumerate(self.components):
            retval.write(str(component))
            if idx < len(self.components) - 1:
                retval.write(" && ")
        retval.write(")")
        return retval.getvalue()

    cdef int _evaluate_vector(self, np.ndarray instantiated_vector):
        cdef ConditionComponent component
        cdef int e

        for i in range(self.n_components):
            component = self.components[i]
            e = component._evaluate_vector(instantiated_vector)
            if e == 0:
                return 0

        return 1

    cdef int _evaluate(self, int I, int* evaluations):
        for i in range(I):
            if evaluations[i] == 0:
                return 0
        return 1


cdef class OrConjunction(AbstractConjunction):
    def __init__(self, *args: AbstractCondition) -> None:
        if len(args) < 2:
            raise ValueError("OrConjunction must at least have two "
                             "Conditions.")
        super(OrConjunction, self).__init__(*args)

    def __repr__(self) -> str:
        retval = io.StringIO()
        retval.write("(")
        for idx, component in enumerate(self.components):
            retval.write(str(component))
            if idx < len(self.components) - 1:
                retval.write(" || ")
        retval.write(")")
        return retval.getvalue()

    cdef int _evaluate(self, int I, int* evaluations):
        for i in range(I):
            if evaluations[i] == 1:
                return 1
        return 0

    cdef int _evaluate_vector(self, np.ndarray instantiated_vector):
        cdef ConditionComponent component
        cdef int e

        for i in range(self.n_components):
            component = self.components[i]
            e = component._evaluate_vector(instantiated_vector)
            if e == 1:
                return 1

        return 0



# Copyright (c) 2014-2016, ConfigSpace developers
# Matthias Feurer
# Katharina Eggensperger
# and others (see commit history).
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from io import StringIO
import os
import tempfile
import unittest

from ConfigSpace.configuration_space import ConfigurationSpace
import ConfigSpace.read_and_write.pcs as pcs
import ConfigSpace.read_and_write.pcs_new as pcs_new
from ConfigSpace.hyperparameters import CategoricalHyperparameter, \
    UniformIntegerHyperparameter, UniformFloatHyperparameter, OrdinalHyperparameter
from ConfigSpace.conditions import EqualsCondition, InCondition, \
    AndConjunction, OrConjunction, NotEqualsCondition, \
    GreaterThanCondition
from ConfigSpace.forbidden import ForbiddenInClause, ForbiddenAndConjunction

# More complex search space
classifier = CategoricalHyperparameter("classifier", ["svm", "nn"])
kernel = CategoricalHyperparameter("kernel", ["rbf", "poly", "sigmoid"])
kernel_condition = EqualsCondition(kernel, classifier, "svm")
C = UniformFloatHyperparameter("C", 0.03125, 32768, log=True)
C_condition = EqualsCondition(C, classifier, "svm")
gamma = UniformFloatHyperparameter("gamma", 0.000030518, 8, log=True)
gamma_condition = EqualsCondition(gamma, kernel, "rbf")
degree = UniformIntegerHyperparameter("degree", 1, 5)
degree_condition = InCondition(degree, kernel, ["poly", "sigmoid"])
neurons = UniformIntegerHyperparameter("neurons", 16, 1024)
neurons_condition = EqualsCondition(neurons, classifier, "nn")
lr = UniformFloatHyperparameter("lr", 0.0001, 1.0)
lr_condition = EqualsCondition(lr, classifier, "nn")
preprocessing = CategoricalHyperparameter("preprocessing", ["None", "pca"])
conditional_space = ConfigurationSpace()
conditional_space.add_hyperparameter(preprocessing)
conditional_space.add_hyperparameter(classifier)
conditional_space.add_hyperparameter(kernel)
conditional_space.add_hyperparameter(C)
conditional_space.add_hyperparameter(neurons)
conditional_space.add_hyperparameter(lr)
conditional_space.add_hyperparameter(degree)
conditional_space.add_hyperparameter(gamma)

conditional_space.add_condition(C_condition)
conditional_space.add_condition(kernel_condition)
conditional_space.add_condition(lr_condition)
conditional_space.add_condition(neurons_condition)
conditional_space.add_condition(degree_condition)
conditional_space.add_condition(gamma_condition)

float_a = UniformFloatHyperparameter("float_a", -1.23, 6.45)
e_float_a = UniformFloatHyperparameter("e_float_a", .5E-2, 4.5e+06)
int_a = UniformIntegerHyperparameter("int_a", -1, 6)
log_a = UniformFloatHyperparameter("log_a", 4e-1, 6.45, log=True)
int_log_a = UniformIntegerHyperparameter("int_log_a", 1, 6, log=True)
cat_a = CategoricalHyperparameter("cat_a", ["a", "b", "c", "d"])
crazy = CategoricalHyperparameter("@.:;/\?!$%&_-<>*+1234567890", ["const"])
easy_space = ConfigurationSpace()
easy_space.add_hyperparameter(float_a)
easy_space.add_hyperparameter(e_float_a)
easy_space.add_hyperparameter(int_a)
easy_space.add_hyperparameter(log_a)
easy_space.add_hyperparameter(int_log_a)
easy_space.add_hyperparameter(cat_a)
easy_space.add_hyperparameter(crazy)


class TestPCSConverter(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_read_configuration_space_basic(self):
        # TODO: what does this test has to do with the PCS converter?
        float_a_copy = UniformFloatHyperparameter("float_a", -1.23, 6.45)
        a_copy = {"a": float_a_copy, "b": int_a}
        a_real = {"b": int_a, "a": float_a}
        self.assertDictEqual(a_real, a_copy)
    
    '''
    Tests for the "older pcs" version
    
    '''
    def test_read_configuration_space_easy(self):
        expected = StringIO()
        expected.write('# This is a \n')
        expected.write('   # This is a comment with a leading whitespace ### ffds \n')
        expected.write('\n')
        expected.write('float_a [-1.23, 6.45] [2.61] # bla\n')
        expected.write('e_float_a [.5E-2, 4.5e+06] [2250000.0025]\n')
        expected.write('int_a [-1, 6] [2]i\n')
        expected.write('log_a [4e-1, 6.45] [1.6062378404]l\n')
        expected.write('int_log_a [1, 6] [2]il\n')
        expected.write('cat_a {a,"b",c,d} [a]\n')
        expected.write('@.:;/\?!$%&_-<>*+1234567890 {"const"} ["const"]\n')
        expected.seek(0)
        cs = pcs.read(expected)
        self.assertEqual(cs, easy_space)
    
    def test_read_configuration_space_conditional(self):
        # More complex search space as string array
        complex_cs = list()
        complex_cs.append("preprocessing {None, pca} [None]")
        complex_cs.append("classifier {svm, nn} [svm]")
        complex_cs.append("kernel {rbf, poly, sigmoid} [rbf]")
        complex_cs.append("C [0.03125, 32768] [32]l")
        complex_cs.append("neurons [16, 1024] [520]i # Should be Q16")
        complex_cs.append("lr [0.0001, 1.0] [0.50005]")
        complex_cs.append("degree [1, 5] [3]i")
        complex_cs.append("gamma [0.000030518, 8] [0.0156251079996]l")

        complex_cs.append("C | classifier in {svm}")
        complex_cs.append("kernel | classifier in {svm}")
        complex_cs.append("lr | classifier in {nn}")
        complex_cs.append("neurons | classifier in {nn}")
        complex_cs.append("degree | kernel in {poly, sigmoid}")
        complex_cs.append("gamma | kernel in {rbf}")

        cs = pcs.read(complex_cs)
        self.assertEqual(cs, conditional_space)

    def test_read_configuration_space_conditional_with_two_parents(self):
        config_space = list()
        config_space.append("@1:0:restarts {F,L,D,x,+,no}[x]")
        config_space.append("@1:S:Luby:aryrestarts {1,2}[1]")
        config_space.append("@1:2:Luby:restarts [1,65535][1000]il")
        config_space.append("@1:2:Luby:restarts | @1:0:restarts in {L}")
        config_space.append("@1:2:Luby:restarts | @1:S:Luby:aryrestarts in {2}")
        cs = pcs.read(config_space)
        self.assertEqual(len(cs.get_conditions()), 1)
        self.assertIsInstance(cs.get_conditions()[0], AndConjunction)

    def test_write_illegal_argument(self):
        sp = {"a": int_a}
        self.assertRaisesRegexp(TypeError, "pcs_parser.write expects an "
                                "instance of "
                                "<class "
                                "'ConfigSpace.configuration_"
                                "space.ConfigurationSpace'>, you provided "
                                "'<(type|class) 'dict'>'", pcs.write, sp)

    def test_write_int(self):
        expected = "int_a [-1, 6] [2]i"
        cs = ConfigurationSpace()
        cs.add_hyperparameter(int_a)
        value = pcs.write(cs)
        self.assertEqual(expected, value)

    def test_write_log_int(self):
        expected = "int_log_a [1, 6] [2]il"
        cs = ConfigurationSpace()
        cs.add_hyperparameter(int_log_a)
        value = pcs.write(cs)
        self.assertEqual(expected, value)

    def test_write_q_int(self):
        expected = "Q16_int_a [16, 1024] [520]i"
        cs = ConfigurationSpace()
        cs.add_hyperparameter(
            UniformIntegerHyperparameter("int_a", 16, 1024, q=16))
        value = pcs.write(cs)
        self.assertEqual(expected, value)

    def test_write_q_float(self):
        expected = "Q16_float_a [16.0, 1024.0] [520.0]"
        cs = ConfigurationSpace()
        cs.add_hyperparameter(
            UniformFloatHyperparameter("float_a", 16, 1024, q=16))
        value = pcs.write(cs)
        self.assertEqual(expected, value)

    def test_write_log10(self):
        expected = "a [10.0, 1000.0] [100.0]l"
        cs = ConfigurationSpace()
        cs.add_hyperparameter(
            UniformFloatHyperparameter("a", 10, 1000, log=True))
        value = pcs.write(cs)
        self.assertEqual(expected, value)

    def test_build_forbidden(self):
        expected = "a {a, b, c} [a]\nb {a, b, c} [c]\n\n" \
                   "{a=a, b=a}\n{a=a, b=b}\n{a=b, b=a}\n{a=b, b=b}"
        cs = ConfigurationSpace()
        a = CategoricalHyperparameter("a", ["a", "b", "c"], "a")
        b = CategoricalHyperparameter("b", ["a", "b", "c"], "c")
        cs.add_hyperparameter(a)
        cs.add_hyperparameter(b)
        fb = ForbiddenAndConjunction(ForbiddenInClause(a, ["a", "b"]),
                                     ForbiddenInClause(b, ["a", "b"]))
        cs.add_forbidden_clause(fb)
        value = pcs.write(cs)
        self.assertIn(expected, value)
    
    """
    Tests for the "newer pcs" version in order to check 
    if both deliver the same results
    """
    def test_read_new_configuration_space_easy(self):
        expected = StringIO()
        expected.write('# This is a \n')
        expected.write('   # This is a comment with a leading whitespace ### ffds \n')
        expected.write('\n')
        expected.write('float_a real [-1.23, 6.45] [2.61] # bla\n')
        expected.write('e_float_a real [.5E-2, 4.5e+06] [2250000.0025]\n')
        expected.write('int_a integer [-1, 6] [2]\n')
        expected.write('log_a real [4e-1, 6.45] [1.6062378404]log\n')
        expected.write('int_log_a integer [1, 6] [2]log\n')
        expected.write('cat_a categorical {a,"b",c,d} [a]\n')
        expected.write('@.:;/\?!$%&_-<>*+1234567890 categorical {"const"} ["const"]\n')
        expected.seek(0)
        cs = pcs_new.read(expected)
        self.assertEqual(cs, easy_space)
        
    def test_read_new_configuration_space_conditional(self):
        # More complex search space as string array
        complex_cs = list()
        complex_cs.append("preprocessing categorical {None, pca} [None]")
        complex_cs.append("classifier categorical {svm, nn} [svm]")
        complex_cs.append("kernel categorical {rbf, poly, sigmoid} [rbf]")
        complex_cs.append("C real [0.03125, 32768] [32]log")
        complex_cs.append("neurons integer [16, 1024] [520] # Should be Q16")
        complex_cs.append("lr real [0.0001, 1.0] [0.50005]")
        complex_cs.append("degree integer [1, 5] [3]")
        complex_cs.append("gamma real [0.000030518, 8] [0.0156251079996]log")

        complex_cs.append("C | classifier in {svm}")
        complex_cs.append("kernel | classifier in {svm}")
        complex_cs.append("lr | classifier in {nn}")
        complex_cs.append("neurons | classifier in {nn}")
        complex_cs.append("degree | kernel in {poly, sigmoid}")
        complex_cs.append("gamma | kernel in {rbf}")

        cs_new = pcs_new.read(complex_cs)
        self.assertEqual(cs_new, conditional_space)
        
        # same in older version
        complex_cs_old = list()
        complex_cs_old.append("preprocessing {None, pca} [None]")
        complex_cs_old.append("classifier {svm, nn} [svm]")
        complex_cs_old.append("kernel {rbf, poly, sigmoid} [rbf]")
        complex_cs_old.append("C [0.03125, 32768] [32]l")
        complex_cs_old.append("neurons [16, 1024] [520]i # Should be Q16")
        complex_cs_old.append("lr [0.0001, 1.0] [0.50005]")
        complex_cs_old.append("degree [1, 5] [3]i")
        complex_cs_old.append("gamma [0.000030518, 8] [0.0156251079996]l")

        complex_cs_old.append("C | classifier in {svm}")
        complex_cs_old.append("kernel | classifier in {svm}")
        complex_cs_old.append("lr | classifier in {nn}")
        complex_cs_old.append("neurons | classifier in {nn}")
        complex_cs_old.append("degree | kernel in {poly, sigmoid}")
        complex_cs_old.append("gamma | kernel in {rbf}")

        cs_old = pcs.read(complex_cs_old)
        self.assertEqual(cs_old, cs_new)
        
    def test_write_new_illegal_argument(self):
        sp = {"a": int_a}
        self.assertRaisesRegexp(TypeError, "pcs_parser.write expects an "
                                "instance of "
                                "<class "
                                "'ConfigSpace.configuration_"
                                "space.ConfigurationSpace'>, you provided "
                                "'<(type|class) 'dict'>'", pcs_new.write, sp)
                                
    def test_write_new_int(self):
        expected = "int_a integer [-1, 6] [2]"
        cs = ConfigurationSpace()
        cs.add_hyperparameter(int_a)
        value = pcs_new.write(cs)
        self.assertEqual(expected, value)

    def test_write_new_log_int(self):
        expected = "int_log_a integer [1, 6] [2]log"
        cs = ConfigurationSpace()
        cs.add_hyperparameter(int_log_a)
        value = pcs_new.write(cs)
        self.assertEqual(expected, value)

    def test_write_new_q_int(self):
        expected = "Q16_int_a integer [16, 1024] [520]"
        cs = ConfigurationSpace()
        cs.add_hyperparameter(
            UniformIntegerHyperparameter("int_a", 16, 1024, q=16))
        value = pcs_new.write(cs)
        self.assertEqual(expected, value)

    def test_write_new_q_float(self):
        expected = "Q16_float_a real [16.0, 1024.0] [520.0]"
        cs = ConfigurationSpace()
        cs.add_hyperparameter(
            UniformFloatHyperparameter("float_a", 16, 1024, q=16))
        value = pcs_new.write(cs)
        self.assertEqual(expected, value)

    def test_write_new_log10(self):
        expected = "a real [10.0, 1000.0] [100.0]log"
        cs = ConfigurationSpace()
        cs.add_hyperparameter(
            UniformFloatHyperparameter("a", 10, 1000, log=True))
        value = pcs_new.write(cs)
        self.assertEqual(expected, value)

    def test_build_new_forbidden(self):
        expected = "a categorical {a, b, c} [a]\nb categorical {a, b, c} [c]\n\n" \
                   "{a=a, b=a}\n{a=a, b=b}\n{a=b, b=a}\n{a=b, b=b}\n"
        cs = ConfigurationSpace()
        a = CategoricalHyperparameter("a", ["a", "b", "c"], "a")
        b = CategoricalHyperparameter("b", ["a", "b", "c"], "c")
        cs.add_hyperparameter(a)
        cs.add_hyperparameter(b)
        fb = ForbiddenAndConjunction(ForbiddenInClause(a, ["a", "b"]),
                                     ForbiddenInClause(b, ["a", "b"]))
        cs.add_forbidden_clause(fb)
        value = pcs_new.write(cs)
        self.assertEqual(expected, value)

    def test_build_new_GreaterThanFloatCondition(self):
        expected = "b integer [0, 10] [5]\n" \
                   "a real [0.0, 1.0] [0.5]\n\n" \
                   "a | b > 5"
        cs = ConfigurationSpace()
        a = UniformFloatHyperparameter("a", 0, 1, 0.5)
        b = UniformIntegerHyperparameter("b", 0, 10, 5)
        cs.add_hyperparameter(a)
        cs.add_hyperparameter(b)
        cond = GreaterThanCondition(a, b, 5)
        cs.add_condition(cond)

        value = pcs_new.write(cs)
        self.assertEqual(expected, value)

        expected = "b real [0.0, 10.0] [5.0]\n" \
                   "a real [0.0, 1.0] [0.5]\n\n" \
                   "a | b > 5"
        cs = ConfigurationSpace()
        a = UniformFloatHyperparameter("a", 0, 1, 0.5)
        b = UniformFloatHyperparameter("b", 0, 10, 5)
        cs.add_hyperparameter(a)
        cs.add_hyperparameter(b)
        cond = GreaterThanCondition(a, b, 5)
        cs.add_condition(cond)

        value = pcs_new.write(cs)
        self.assertEqual(expected, value)

    def test_build_new_GreaterThanIntCondition(self):
        expected = "a real [0.0, 1.0] [0.5]\n" \
                   "b integer [0, 10] [5]\n\n" \
                   "b | a > 0.5"
        cs = ConfigurationSpace()
        a = UniformFloatHyperparameter("a", 0, 1, 0.5)
        b = UniformIntegerHyperparameter("b", 0, 10, 5)
        cs.add_hyperparameter(a)
        cs.add_hyperparameter(b)
        cond = GreaterThanCondition(b, a, 0.5)
        cs.add_condition(cond)

        value = pcs_new.write(cs)
        self.assertEqual(expected, value)

        expected = "a integer [0, 10] [5]\n" \
                   "b integer [0, 10] [5]\n\n" \
                   "b | a > 5"
        cs = ConfigurationSpace()
        a = UniformIntegerHyperparameter("a", 0, 10, 5)
        b = UniformIntegerHyperparameter("b", 0, 10, 5)
        cs.add_hyperparameter(a)
        cs.add_hyperparameter(b)
        cond = GreaterThanCondition(b, a, 5)
        cs.add_condition(cond)

        value = pcs_new.write(cs)
        self.assertEqual(expected, value)

    def test_read_new_configuration_space_complex_conditionals(self):
        classi = OrdinalHyperparameter("classi", ["random_forest","extra_trees","k_nearest_neighbors","something"])
        knn_weights = CategoricalHyperparameter("knn_weights", ["uniform", "distance"])
        weather = OrdinalHyperparameter("weather",["sunny", "rainy", "cloudy", "snowing"])
        temperature = CategoricalHyperparameter("temperature", ["high", "low"])
        rain = CategoricalHyperparameter("rain", ["yes", "no"])
        gloves = OrdinalHyperparameter("gloves",["none", "yarn", "leather", "gortex"])
        heur1 = CategoricalHyperparameter("heur1", ["off", "on"])
        heur2 = CategoricalHyperparameter("heur2", ["off", "on"])
        heur_order = CategoricalHyperparameter("heur_order", ["heur1then2", "heur2then1"])
        gloves_condition = OrConjunction(EqualsCondition(gloves, rain, "yes"),
                                           EqualsCondition(gloves, temperature, "low"))
        heur_condition = AndConjunction(EqualsCondition(heur_order, heur1, "on"),
                                        EqualsCondition(heur_order, heur2, "on"))
        and_conjunction = AndConjunction(NotEqualsCondition(knn_weights, classi, "extra_trees"), 
                                         EqualsCondition(knn_weights, classi, "random_forest"))
        Cl_condition = OrConjunction(EqualsCondition(knn_weights, classi, "k_nearest_neighbors"),
                                     and_conjunction,
                                     EqualsCondition(knn_weights, classi, "something"))
        
        and1 = AndConjunction(EqualsCondition(temperature, weather, "rainy"),
                              EqualsCondition(temperature, weather, "cloudy"))   
        and2 = AndConjunction(EqualsCondition(temperature, weather, "sunny"),
                              NotEqualsCondition(temperature, weather, "snowing"))                       
        another_condition = OrConjunction(and1,and2)
        
        complex_conditional_space = ConfigurationSpace()
        complex_conditional_space.add_hyperparameter(classi)
        complex_conditional_space.add_hyperparameter(knn_weights)
        complex_conditional_space.add_hyperparameter(weather)
        complex_conditional_space.add_hyperparameter(temperature)
        complex_conditional_space.add_hyperparameter(rain)
        complex_conditional_space.add_hyperparameter(gloves)
        complex_conditional_space.add_hyperparameter(heur1)
        complex_conditional_space.add_hyperparameter(heur2)
        complex_conditional_space.add_hyperparameter(heur_order)
        
        complex_conditional_space.add_condition(gloves_condition)
        complex_conditional_space.add_condition(heur_condition)
        complex_conditional_space.add_condition(Cl_condition)
        complex_conditional_space.add_condition(another_condition)

        complex_cs = list()
        complex_cs.append("classi ordinal {random_forest,extra_trees,k_nearest_neighbors, something} [random_forest]")
        complex_cs.append("knn_weights categorical {uniform, distance} [uniform]")
        complex_cs.append("weather ordinal {sunny, rainy, cloudy, snowing} [sunny]")
        complex_cs.append("temperature categorical {high, low} [high]")
        complex_cs.append("rain categorical { yes, no } [yes]")
        complex_cs.append("gloves ordinal { none, yarn, leather, gortex } [none]")
        complex_cs.append("heur1 categorical { off, on } [off]")
        complex_cs.append("heur2 categorical { off, on } [off]")
        complex_cs.append("heur_order categorical { heur1then2, heur2then1 } [heur1then2]")
        complex_cs.append("gloves | rain == yes || temperature == low")
        complex_cs.append("heur_order | heur1 == on && heur2 == on")
        complex_cs.append("knn_weights | classi == k_nearest_neighbors || classi != extra_trees && classi == random_forest || classi == something")
        complex_cs.append("temperature | weather == rainy && weather == cloudy || weather == sunny && weather != snowing")
        cs_new = pcs_new.read(complex_cs)
        self.assertEqual(cs_new, complex_conditional_space)

    def test_convert_restrictions(self):
        # This is a smoke test to make sure that the int/float values in the
        # greater or smaller statements are converted to the right type when
        # reading them
        s = """x1 real [0,1] [0]
        x2 real [0,1] [0]
        x3 real [0,1] [0]
        x4 integer [0,2] [0]
        x5 real [0,1] [0]
        x6 ordinal {cold, luke-warm, hot} [cold]
        x1 | x2 > 0.5
        x3 | x4 > 1 && x4 == 2 && x4 in {2}
        x5 | x6 > luke-warm"""

        pcs_new.read(s.split('\n'))

    def test_write_restrictions(self):
        s = "c integer [0, 2] [0]\n" + \
            "d ordinal {cold, luke-warm, hot} [cold]\n" + \
            "e real [0.0, 1.0] [0.0]\n" + \
            "b real [0.0, 1.0] [0.0]\n" + \
            "a real [0.0, 1.0] [0.0]\n" + \
            "\n" + \
            "b | d in {luke-warm, hot} || c > 1\n" + \
            "a | b == 0.5 && e > 0.5"

        a = pcs_new.read(s.split('\n'))
        out = pcs_new.write(a)
        self.assertEqual(out, s)

    def test_read_write(self):
        # Some smoke tests whether reading, writing, reading alters makes the
        #  configspace incomparable
        this_file = os.path.abspath(__file__)
        this_directory = os.path.dirname(this_file)
        configuration_space_path = os.path.join(this_directory,
                                                "..", "test_searchspaces")
        configuration_space_path = os.path.abspath(configuration_space_path)
        configuration_space_path = os.path.join(configuration_space_path,
                                                "spear-params-mixed.pcs")
        with open(configuration_space_path) as fh:
            cs = pcs.read(fh)

        tf = tempfile.NamedTemporaryFile()
        name = tf.name
        tf.close()
        with open(name, 'w') as fh:
            pcs_string = pcs.write(cs)
            fh.write(pcs_string)
        with open(name, 'r') as fh:
            pcs_new = pcs.read(fh)

        self.assertEqual(pcs_new, cs, msg=(pcs_new, cs))




# Copyright (c) 2014-2016, ConfigSpace developers
# Matthias Feurer
# Katharina Eggensperger
# and others (see commit history).
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# __authors__ = ["Katharina Eggensperger", "Matthias Feurer"]
# __contact__ = "automl.org"
#
# import unittest
#
# import six
#
# import HPOlibConfigSpace.converters.pb_parser as pb_parser
# from HPOlibConfigSpace.hyperparameters import UniformFloatHyperparameter, \
#     UniformIntegerHyperparameter, CategoricalHyperparameter
#
# float_a = UniformFloatHyperparameter("float_a", -5.3, 10.5)
# int_a = UniformIntegerHyperparameter("int_a", -5, 10)
# cat_a = CategoricalHyperparameter("enum_a", ["5", "a", "b", "@/&%%"])
# crazy = CategoricalHyperparameter("@.:;/\?!$%&_-<>*+1234567890", ["const"])
# easy_space = {"float_a": float_a,
#               "int_a": int_a,
#               "enum_a": cat_a,
#               "@.:;/\?!$%&_-<>*+1234567890": crazy,
#               }
#
#
# class TestPbConverter(unittest.TestCase):
#
#     def test_read_configuration_space_easy(self):
#         expected = StringIO.StringIO()
#         expected.write("""
#         language: PYTHON
#         name: "HPOlib.cv"
#
#         variable {
#             name: "float_a"
#             type: FLOAT
#             size: 1
#             min:  -5.3
#             max:  10.5
#         }
#
#         variable {
#             name: "enum_a"
#             type: ENUM
#             size: 1
#             options: "5"
#             options: "a"
#             options: "b"
#             options: "@/&%%"
#         }
#
#          variable {
#             name: "int_a"
#             type: INT
#             size: 1
#             min:  -5
#             max:  10
#          }
#
#          variable {
#             name: "@.:;/\?!$%&_-<>*+1234567890"
#             type: ENUM
#             size: 1
#             options: "const"
#         }
#         """)
#
#         cs = pb_parser.read(expected.getvalue())
#         self.assertDictEqual(cs, easy_space)



import os
import re
import unittest

from ConfigSpace.read_and_write.json import read, write
from ConfigSpace.read_and_write.pcs import read as read_pcs
from ConfigSpace.read_and_write.pcs_new import read as read_pcs_new


class TestJson(unittest.TestCase):

    def test_round_trip(self):
        this_file = os.path.abspath(__file__)
        this_directory = os.path.dirname(this_file)
        configuration_space_path = os.path.join(this_directory,
                                                "..", "test_searchspaces")
        configuration_space_path = os.path.abspath(configuration_space_path)
        pcs_files = os.listdir(configuration_space_path)

        for pcs_file in sorted(pcs_files):

            if '.pcs' in pcs_file:
                full_path = os.path.join(configuration_space_path, pcs_file)

                with open(full_path) as fh:
                    cs_string = fh.read().split('\n')
                try:
                    cs = read_pcs(cs_string)
                except:
                    cs = read_pcs_new(cs_string)

                cs.name = pcs_file

                json_string = write(cs)
                new_cs = read(json_string)

                self.assertEqual(new_cs, cs)


# raise NotImplementedError()
# #!/usr/bin/env python
#
# ##
# # wrapping: A program making it easy to use hyperparameter
# # optimization software.
# # Copyright (C) 2013 Katharina Eggensperger and Matthias Feurer
# #
# # This program is free software: you can redistribute it and/or modify
# # it under the terms of the GNU General Public License as published by
# # the Free Software Foundation, either version 3 of the License, or
# # (at your option) any later version.
# #
# # This program is distributed in the hope that it will be useful,
# # but WITHOUT ANY WARRANTY; without even the implied warranty of
# # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# # GNU General Public License for more details.
# #
# # You should have received a copy of the GNU General Public License
# # along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# __authors__ = ["Katharina Eggensperger", "Matthias Feurer"]
# __contact__ = "automl.org"
#
# import sys
# from google.protobuf import text_format
# import numpy as np
#
# from HPOlibConfigSpace.converters.spearmint_april2013_mod_spearmint_pb2 import Experiment, PYTHON
# import HPOlibConfigSpace.configuration_space as configuration_space
# from HPOlibConfigSpace.hyperparameters import UniformFloatHyperparameter, \
#     UniformIntegerHyperparameter, CategoricalHyperparameter
#
#
# def read(pb_string):
#     searchspace = dict()
#     if type(pb_string) == file:
#         pb_string = pb_string.read()
#
#     if not isinstance(pb_string, str):
#         raise ValueError("Input is not a string or a file")
#
#     pb_space = Experiment()
#     text_format.Merge(pb_string, pb_space)
#     for para in pb_space.variable:
#         name = str(para.name)
#         size = para.size
#         if size != 1:
#             raise NotImplementedError("%s has size %s, we only convert with size 1" % (name, size))
#         if para.type == Experiment.ParameterSpec.ENUM:
#             choices = list()
#             for ch in para.options:
#                 choices.append(str(ch))
#             searchspace[name] = CategoricalHyperparameter(
#                 name=name, choices=choices, conditions=None)
#         elif para.type == Experiment.ParameterSpec.FLOAT:
#             searchspace[name] = UniformFloatHyperparameter(
#                 name=name, lower=para.min, upper=para.max, base=None, q=None, conditions=None)
#         elif para.type == Experiment.ParameterSpec.INT:
#             searchspace[name] = UniformIntegerHyperparameter(
#                 name=name, lower=para.min, upper=para.max, base=None, q=None, conditions=None)
#         else:
#             raise NotImplementedError("Don't know that type: %s (%s)" %
#                                       (type(para), para.name))
#     return searchspace
#
#
# def write(searchspace):
#     exp = Experiment()
#     # Assumption: Algo is written in Python called cv.py, only works for HPOlib
#     exp.language = PYTHON
#     exp.name = "HPOlib.cv"
#
#     tmp_para_list = list()
#     for parakey in searchspace:
#         param = searchspace[parakey]
#         # We use an Experiment to store the variables
#         constructed_param = Experiment.ParameterSpec()
#         constructed_param.size = 1
#         constructed_param.name = param.name
#         if isinstance(param, configuration_space.CategoricalHyperparameter):
#             constructed_param.type = Experiment.ParameterSpec.ENUM
#             for choice in param.choices:
#                 constructed_param.options.append(choice)
#         else:
#             if isinstance(param, configuration_space.IntegerHyperparameter):
#                 constructed_param.type = Experiment.ParameterSpec.INT
#                 constructed_param.min = param.lower
#                 constructed_param.max = param.upper
#             elif isinstance(param, configuration_space.FloatHyperparameter):
#                 constructed_param.type = Experiment.ParameterSpec.FLOAT
#                 constructed_param.min = param.lower
#                 constructed_param.max = param.upper
#             else:
#                 raise NotImplementedError("Unknown type: %s (%s)" %
#                                           (type(param), param.name))
#
#             # Handle LOG params
#             if param.base is not None:
#                 if int(param.base) != param.base:
#                     raise NotImplementedError("We cannot handle non-int bases: %s (%s)" %
#                                               (str(param.base), param.name))
#                 constructed_param.name = "LOG%d_%s" % (int(param.base), constructed_param.name)
#                 constructed_param.min = np.log10(constructed_param.min) / np.log10(param.base)
#                 constructed_param.max = np.log10(constructed_param.max) / np.log10(param.base)
#
#             # Handle q params
#             if param.q is not None:
#                 constructed_param.name = "Q%d_%s" % (int(param.q), constructed_param.name)
#
#             # Do NOT handle conditions:
#             if param.has_conditions():
#                 print("WARNING: We loose condition: %s" % str(param.conditions))
#         tmp_para_list.append(constructed_param)
#     exp.variable.extend(tmp_para_list)
#
#     result = text_format.MessageToString(exp)
#     return result
#
#
# if __name__ == "__main__":
#     fh = open(sys.argv[1])
#     sp = read(fh)
#     fh.close()
#     print("==== READ Result")
#     print("\n".join(["%s %s" % (i, sp[i]) for i in sp]))
#
#     print("==== WRITE Result")
#     created_sp = write(sp)
#     print(created_sp)


# Copyright (c) 2014-2016, ConfigSpace developers
# Matthias Feurer
# Katharina Eggensperger
# and others (see commit history).
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from itertools import product
import unittest
import warnings

import numpy as np

from ConfigSpace.hyperparameters import \
    UniformIntegerHyperparameter, CategoricalHyperparameter

# from ConfigSpace.forbidden import ForbiddenEqualsClause, \
#     ForbiddenInClause, ForbiddenAndConjunction
from ConfigSpace.forbidden import ForbiddenEqualsClause, \
    ForbiddenInClause, ForbiddenAndConjunction


class TestForbidden(unittest.TestCase):
    # TODO: return only copies of the objects!

    def test_forbidden_equals_clause(self):
        hp1 = CategoricalHyperparameter("parent", [0, 1])
        hp2 = UniformIntegerHyperparameter("child", 0, 10)
        hp3 = CategoricalHyperparameter("grandchild", ["hot", "cold"])

        self.assertRaisesRegexp(
            TypeError,
            "Argument 'hyperparameter' has incorrect type \(expected ConfigSpace.hyperparameters.Hyperparameter, got str\)",
            ForbiddenEqualsClause, "HP1", 1,
        )

        self.assertRaisesRegexp(
            ValueError,
            "Forbidden clause must be instantiated with a legal hyperparameter value for "
            "'parent, Type: Categorical, Choices: \{0, 1\}, Default: 0', but got '2'",
            ForbiddenEqualsClause, hp1, 2,
        )

        forb1 = ForbiddenEqualsClause(hp1, 1)
        forb1_ = ForbiddenEqualsClause(hp1, 1)
        forb1__ = ForbiddenEqualsClause(hp1, 0)
        forb2 = ForbiddenEqualsClause(hp2, 10)
        forb3 = ForbiddenEqualsClause(hp3, "hot")
        forb3_ = ForbiddenEqualsClause(hp3, "hot")

        self.assertEqual(forb3, forb3_)
        # print("\eq0:", 1, 1)
        # self.assertEqual(1, 1)
        # print("\neq1:", forb1, forb1_)
        self.assertEqual(forb1, forb1_)
        # print("\nneq2:", forb1, "forb1")
        self.assertNotEqual(forb1, "forb1")
        # print("\nneq3:", forb1, forb2)
        self.assertNotEqual(forb1, forb2)
        # print("\nneq4:", forb1_, forb1)
        self.assertNotEqual(forb1__, forb1)
        # print("\neq5:", "Forbidden: parent == 1", str(forb1))
        self.assertEqual("Forbidden: parent == 1", str(forb1))

        # print("\nraisereg6:")
        self.assertRaisesRegexp(ValueError,
                                "Is_forbidden must be called with the "
                                "instanstatiated hyperparameter in the "
                                "forbidden clause; you are missing "
                                "'parent'", forb1.is_forbidden,
                                {1: hp2}, True)
        # print("\nneq7:")
        self.assertFalse(forb1.is_forbidden({'child': 1}, strict=False))
        # print("\nneq8:")
        self.assertFalse(forb1.is_forbidden({'parent': 0}, True))
        # print("\nneq9:")
        self.assertTrue(forb1.is_forbidden({'parent': 1}, True))

        # Test forbidden on vector values
        hyperparameter_idx = {
            hp1.name: 0,
            hp2.name: 1
        }
        forb1.set_vector_idx(hyperparameter_idx)
        # print("\nneq10:")
        self.assertFalse(forb1.is_forbidden_vector(np.array([np.NaN, np.NaN]), strict=False))
        # print("\nneq11:")
        self.assertFalse(forb1.is_forbidden_vector(np.array([0., np.NaN]), strict=False))
        # print("\nneq12:")
        self.assertTrue(forb1.is_forbidden_vector(np.array([1., np.NaN]), strict=False))


    def test_in_condition(self):
        hp1 = CategoricalHyperparameter("parent", [0, 1, 2, 3, 4])
        hp2 = UniformIntegerHyperparameter("child", 0, 10)
        hp3 = UniformIntegerHyperparameter("child2", 0, 10)
        hp4 = CategoricalHyperparameter("grandchild", ["hot", "cold", "warm"])

        self.assertRaisesRegexp(
            TypeError,
            "Argument 'hyperparameter' has incorrect type \(expected ConfigSpace.hyperparameters.Hyperparameter, got str\)",
            ForbiddenInClause, "HP1", 1,
        )

        self.assertRaisesRegexp(
            ValueError,
            "Forbidden clause must be instantiated with a "
            "legal hyperparameter value for "
            "'parent, Type: Categorical, Choices: {0, 1, 2, 3, 4}, "
            "Default: 0', but got '5'",
            ForbiddenInClause, hp1, [5],
        )

        forb1 = ForbiddenInClause(hp2, [5, 6, 7, 8, 9])
        forb1_ = ForbiddenInClause(hp2, [9, 8, 7, 6, 5])
        forb2 = ForbiddenInClause(hp2, [5, 6, 7, 8])
        forb3 = ForbiddenInClause(hp3, [5, 6, 7, 8, 9])
        forb4 = ForbiddenInClause(hp4, ["hot", "cold"])
        forb4_ = ForbiddenInClause(hp4, ["hot", "cold"])
        forb5 = ForbiddenInClause(hp1, [3, 4])
        forb5_ = ForbiddenInClause(hp1, [3, 4])

        self.assertEqual(forb5, forb5_)
        self.assertEqual(forb4, forb4_)

        # print("\nTest1:")
        self.assertEqual(forb1, forb1_)
        # print("\nTest2:")
        self.assertNotEqual(forb1, forb2)
        # print("\nTest3:")
        self.assertNotEqual(forb1, forb3)
        # print("\nTest4:")
        self.assertEqual("Forbidden: child in {5, 6, 7, 8, 9}", str(forb1))
        # print("\nTest5:")
        self.assertRaisesRegexp(ValueError,
                                "Is_forbidden must be called with the "
                                "instanstatiated hyperparameter in the "
                                "forbidden clause; you are missing "
                                "'child'", forb1.is_forbidden,
                                {'parent': 1}, True)
        # print("\nTest6:")
        self.assertFalse(forb1.is_forbidden({'parent': 1}, strict=False))
        # print("\nTest7:")
        for i in range(0, 5):
            self.assertFalse(forb1.is_forbidden({'child': i}, True))
        # print("\nTest8:")
        for i in range(5, 10):
            self.assertTrue(forb1.is_forbidden({'child': i}, True))

        # Test forbidden on vector values
        hyperparameter_idx = {
            hp1.name: 0,
            hp2.name: 1
        }
        forb1.set_vector_idx(hyperparameter_idx)
        # print("\nTest9:")
        self.assertFalse(forb1.is_forbidden_vector(np.array([np.NaN, np.NaN]), strict=False))
        # print("\nTest10:")
        self.assertFalse(forb1.is_forbidden_vector(np.array([np.NaN, 0]), strict=False))
        correct_vector_value = hp2._inverse_transform(6)
        # print("\nTest11:")
        print(correct_vector_value, np.array([np.NaN, correct_vector_value]))
        self.assertTrue(forb1.is_forbidden_vector(np.array([np.NaN, correct_vector_value]), strict=False))
        # print("\nfinished:")

    def test_and_conjunction(self):
        hp1 = CategoricalHyperparameter("parent", [0, 1])
        hp2 = UniformIntegerHyperparameter("child", 0, 2)
        hp3 = UniformIntegerHyperparameter("child2", 0, 2)
        hp4 = UniformIntegerHyperparameter("child3", 0, 2)

        forb2 = ForbiddenEqualsClause(hp1, 1)
        forb3 = ForbiddenInClause(hp2, range(2, 3))
        forb4 = ForbiddenInClause(hp3, range(2, 3))
        forb5 = ForbiddenInClause(hp4, range(2, 3))

        and1 = ForbiddenAndConjunction(forb2, forb3)
        and2 = ForbiddenAndConjunction(forb2, forb4)
        and3 = ForbiddenAndConjunction(forb2, forb5)

        total_and = ForbiddenAndConjunction(and1, and2, and3)
        self.assertEqual("((Forbidden: parent == 1 && Forbidden: child in {2}) "
                         "&& (Forbidden: parent == 1 && Forbidden: child2 in {2}) "
                         "&& (Forbidden: parent == 1 && Forbidden: child3 in "
                         "{2}))", str(total_and))

        results = [False, False, False, False, False,
                   False, False, False, False, False,
                   False, False, False, False, False,
                   False, False, False, False, False,
                   False, False, False, False, False,
                   False, False, False, False, False,
                   False, False, False, False, False,
                   False, False, False, False, False,
                   False, False, False, False, False,
                   False, False, False, False, False,
                   False, False, False, True]

        for i, values in enumerate(product(range(2), range(3), range(3),
                                           range(3))):
            is_forbidden = total_and.is_forbidden(
                {"parent": values[0],
                 "child": values[1],
                 "child2": values[2],
                 "child3": values[3]},
                True,
            )

            self.assertEqual(results[i], is_forbidden)

            self.assertFalse(total_and.is_forbidden({}, strict=False))



##
# Copyright (C) 2012 Jasper Snoek, Hugo Larochelle and Ryan P. Adams
# 
# This code is written for research and educational purposes only to 
# supplement the paper entitled
# "Practical Bayesian Optimization of Machine Learning Algorithms"
# by Snoek, Larochelle and Adams
# Advances in Neural Information Processing Systems, 2012
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
##
# Copyright (C) 2012 Jasper Snoek, Hugo Larochelle and Ryan P. Adams
# 
# This code is written for research and educational purposes only to 
# supplement the paper entitled
# "Practical Bayesian Optimization of Machine Learning Algorithms"
# by Snoek, Larochelle and Adams
# Advances in Neural Information Processing Systems, 2012
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Generated by the protocol buffer compiler.  DO NOT EDIT!

from google.protobuf import descriptor
from google.protobuf import message
from google.protobuf import reflection
# from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

DESCRIPTOR = descriptor.FileDescriptor(
    name='spearmint.proto',
    package='',
    serialized_pb="\n\x0fspearmint.proto\"\xcc\x01\n\x03Job\x12\n\n\x02id\x18\x01 \x02(\x04\x12\x10\n\x08\x65xpt_dir" +
                  "\x18\x02 \x02(\t\x12\x0c\n\x04name\x18\x03 \x02(\t\x12\x1b\n\x08language\x18\x04 \x02" +
                  "(\x0e\x32\t.Language\x12\x0e\n\x06status\x18\x05 \x01(\t\x12\x19\n\x05param\x18\x06 \x03" +
                  "(\x0b\x32\n.Parameter\x12\x10\n\x08submit_t\x18\x07 \x01(\x04\x12\x0f\n\x07start_t\x18\x08 " +
                  "\x01(\x04\x12\r\n\x05\x65nd_t\x18\t \x01(\x04\x12\r\n\x05value\x18\n \x01(\x01\x12\x10\n\x08\x64" +
                  "uration\x18\x0b \x01(\x01\"L\n\tParameter\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\x0f\n\x07int_val" +
                  "\x18\x02 \x03(\x03\x12\x0f\n\x07str_val\x18\x03 \x03(\t\x12\x0f\n\x07\x64\x62l_val\x18\x04 \x03" +
                  "(\x01\"\x91\x02\n\nExperiment\x12\x1b\n\x08language\x18\x01 \x02(\x0e\x32\t.Language\x12\x0c\n\x04" +
                  "name\x18\x02 \x02(\t\x12+\n\x08variable\x18\x03 \x03(\x0b\x32\x19.Experiment.ParameterSpec" +
                  "\x1a\xaa\x01\n\rParameterSpec\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\x0c\n\x04size\x18\x02 \x02" +
                  "(\r\x12,\n\x04type\x18\x03 \x02(\x0e\x32\x1e.Experiment.ParameterSpec.Type\x12\x0f\n\x07options" +
                  "\x18\x04 \x03(\t\x12\x0b\n\x03min\x18\x05 \x01(\x01\x12\x0b\n\x03max\x18\x06 \x01(\x01\"$\n\x04" +
                  "Type\x12\x07\n\x03INT\x10\x01\x12\t\n\x05\x46LOAT\x10\x02\x12\x08\n\x04\x45NUM\x10\x03*6\n\x08" +
                  "Language\x12\n\n\x06MATLAB\x10\x01\x12\n\n\x06PYTHON\x10\x02\x12\t\n\x05SHELL\x10\x03\x12\x07" +
                  "\n\x03MCR\x10\x04")

_LANGUAGE = descriptor.EnumDescriptor(
    name='Language',
    full_name='Language',
    filename=None,
    file=DESCRIPTOR,
    values=[
        descriptor.EnumValueDescriptor(
            name='MATLAB', index=0, number=1,
            options=None,
            type=None),
        descriptor.EnumValueDescriptor(
            name='PYTHON', index=1, number=2,
            options=None,
            type=None),
        descriptor.EnumValueDescriptor(
            name='SHELL', index=2, number=3,
            options=None,
            type=None),
        descriptor.EnumValueDescriptor(
            name='MCR', index=3, number=4,
            options=None,
            type=None),
    ],
    containing_type=None,
    options=None,
    serialized_start=580,
    serialized_end=634,
)

MATLAB = 1
PYTHON = 2
SHELL = 3
MCR = 4


_EXPERIMENT_PARAMETERSPEC_TYPE = descriptor.EnumDescriptor(
    name='Type',
    full_name='Experiment.ParameterSpec.Type',
    filename=None,
    file=DESCRIPTOR,
    values=[
        descriptor.EnumValueDescriptor(
            name='INT', index=0, number=1,
            options=None,
            type=None),
        descriptor.EnumValueDescriptor(
            name='FLOAT', index=1, number=2,
            options=None,
            type=None),
        descriptor.EnumValueDescriptor(
            name='ENUM', index=2, number=3,
            options=None,
            type=None),
    ],
    containing_type=None,
    options=None,
    serialized_start=542,
    serialized_end=578,
)


_JOB = descriptor.Descriptor(
    name='Job',
    full_name='Job',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        descriptor.FieldDescriptor(
            name='id', full_name='Job.id', index=0,
            number=1, type=4, cpp_type=4, label=2,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        descriptor.FieldDescriptor(
            name='expt_dir', full_name='Job.expt_dir', index=1,
            number=2, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=unicode("", "utf-8"),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        descriptor.FieldDescriptor(
            name='name', full_name='Job.name', index=2,
            number=3, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=unicode("", "utf-8"),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        descriptor.FieldDescriptor(
            name='language', full_name='Job.language', index=3,
            number=4, type=14, cpp_type=8, label=2,
            has_default_value=False, default_value=1,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        descriptor.FieldDescriptor(
            name='status', full_name='Job.status', index=4,
            number=5, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=unicode("", "utf-8"),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        descriptor.FieldDescriptor(
            name='param', full_name='Job.param', index=5,
            number=6, type=11, cpp_type=10, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        descriptor.FieldDescriptor(
            name='submit_t', full_name='Job.submit_t', index=6,
            number=7, type=4, cpp_type=4, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        descriptor.FieldDescriptor(
            name='start_t', full_name='Job.start_t', index=7,
            number=8, type=4, cpp_type=4, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        descriptor.FieldDescriptor(
            name='end_t', full_name='Job.end_t', index=8,
            number=9, type=4, cpp_type=4, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        descriptor.FieldDescriptor(
            name='value', full_name='Job.value', index=9,
            number=10, type=1, cpp_type=5, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        descriptor.FieldDescriptor(
            name='duration', full_name='Job.duration', index=10,
            number=11, type=1, cpp_type=5, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    extension_ranges=[],
    serialized_start=20,
    serialized_end=224,
)


_PARAMETER = descriptor.Descriptor(
    name='Parameter',
    full_name='Parameter',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        descriptor.FieldDescriptor(
            name='name', full_name='Parameter.name', index=0,
            number=1, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=unicode("", "utf-8"),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        descriptor.FieldDescriptor(
            name='int_val', full_name='Parameter.int_val', index=1,
            number=2, type=3, cpp_type=2, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        descriptor.FieldDescriptor(
            name='str_val', full_name='Parameter.str_val', index=2,
            number=3, type=9, cpp_type=9, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        descriptor.FieldDescriptor(
            name='dbl_val', full_name='Parameter.dbl_val', index=3,
            number=4, type=1, cpp_type=5, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    extension_ranges=[],
    serialized_start=226,
    serialized_end=302,
)


_EXPERIMENT_PARAMETERSPEC = descriptor.Descriptor(
    name='ParameterSpec',
    full_name='Experiment.ParameterSpec',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        descriptor.FieldDescriptor(
            name='name', full_name='Experiment.ParameterSpec.name', index=0,
            number=1, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=unicode("", "utf-8"),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        descriptor.FieldDescriptor(
            name='size', full_name='Experiment.ParameterSpec.size', index=1,
            number=2, type=13, cpp_type=3, label=2,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        descriptor.FieldDescriptor(
            name='type', full_name='Experiment.ParameterSpec.type', index=2,
            number=3, type=14, cpp_type=8, label=2,
            has_default_value=False, default_value=1,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        descriptor.FieldDescriptor(
            name='options', full_name='Experiment.ParameterSpec.options', index=3,
            number=4, type=9, cpp_type=9, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        descriptor.FieldDescriptor(
            name='min', full_name='Experiment.ParameterSpec.min', index=4,
            number=5, type=1, cpp_type=5, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        descriptor.FieldDescriptor(
            name='max', full_name='Experiment.ParameterSpec.max', index=5,
            number=6, type=1, cpp_type=5, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[_EXPERIMENT_PARAMETERSPEC_TYPE, ],
    options=None,
    is_extendable=False,
    extension_ranges=[],
    serialized_start=408,
    serialized_end=578,
)

_EXPERIMENT = descriptor.Descriptor(
    name='Experiment',
    full_name='Experiment',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        descriptor.FieldDescriptor(
            name='language', full_name='Experiment.language', index=0,
            number=1, type=14, cpp_type=8, label=2,
            has_default_value=False, default_value=1,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        descriptor.FieldDescriptor(
            name='name', full_name='Experiment.name', index=1,
            number=2, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=unicode("", "utf-8"),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        descriptor.FieldDescriptor(
            name='variable', full_name='Experiment.variable', index=2,
            number=3, type=11, cpp_type=10, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
    ],
    extensions=[],
    nested_types=[_EXPERIMENT_PARAMETERSPEC, ],
    enum_types=[],
    options=None,
    is_extendable=False,
    extension_ranges=[],
    serialized_start=305,
    serialized_end=578,
)

_JOB.fields_by_name['language'].enum_type = _LANGUAGE
_JOB.fields_by_name['param'].message_type = _PARAMETER
_EXPERIMENT_PARAMETERSPEC.fields_by_name['type'].enum_type = _EXPERIMENT_PARAMETERSPEC_TYPE
_EXPERIMENT_PARAMETERSPEC.containing_type = _EXPERIMENT;
_EXPERIMENT_PARAMETERSPEC_TYPE.containing_type = _EXPERIMENT_PARAMETERSPEC;
_EXPERIMENT.fields_by_name['language'].enum_type = _LANGUAGE
_EXPERIMENT.fields_by_name['variable'].message_type = _EXPERIMENT_PARAMETERSPEC
DESCRIPTOR.message_types_by_name['Job'] = _JOB
DESCRIPTOR.message_types_by_name['Parameter'] = _PARAMETER
DESCRIPTOR.message_types_by_name['Experiment'] = _EXPERIMENT


class Job(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _JOB

    # @@protoc_insertion_point(class_scope:Job)


class Parameter(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _PARAMETER

    # @@protoc_insertion_point(class_scope:Parameter)


class Experiment(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType

    class ParameterSpec(message.Message):
        __metaclass__ = reflection.GeneratedProtocolMessageType
        DESCRIPTOR = _EXPERIMENT_PARAMETERSPEC

    # @@protoc_insertion_point(class_scope:Experiment.ParameterSpec)
    DESCRIPTOR = _EXPERIMENT

    # @@protoc_insertion_point(class_scope:Experiment)

# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python

##
# wrapping: A program making it easy to use hyperparameter
# optimization software.
# Copyright (C) 2013 Katharina Eggensperger and Matthias Feurer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__authors__ = ["Katharina Eggensperger", "Matthias Feurer", "Mohsin Ali"]
__contact__ = "automl.org"

from collections import OrderedDict
from itertools import product
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import CategoricalHyperparameter, \
    UniformIntegerHyperparameter, UniformFloatHyperparameter, \
    NumericalHyperparameter, Constant, IntegerHyperparameter, \
    NormalIntegerHyperparameter, NormalFloatHyperparameter, OrdinalHyperparameter
from ConfigSpace.conditions import EqualsCondition, NotEqualsCondition, \
    InCondition, AndConjunction, OrConjunction, ConditionComponent, GreaterThanCondition, LessThanCondition
# from ConfigSpace.forbidden import ForbiddenEqualsClause, \
#     ForbiddenAndConjunction, ForbiddenInClause, AbstractForbiddenComponent, MultipleValueForbiddenClause
from ConfigSpace.forbidden import ForbiddenEqualsClause, \
    ForbiddenAndConjunction, ForbiddenInClause, AbstractForbiddenComponent, MultipleValueForbiddenClause
import sys
import pyparsing
import io
import numpy as np

# Build pyparsing expressions for params
pp_param_name = pyparsing.Word(pyparsing.alphanums + "_" + "-" + "@" + "." + ":" + ";" + "\\" + "/" + "?" + "!" +
                               "$" + "%" + "&" + "*" + "+" + "<" + ">")

def build_categorical(param):
    cat_template = "%s '--%s ' c {%s}"
    return cat_template % (param.name, param.name,
                           ",".join([str(value) for value in param.choices]))

def build_ordinal(param):
    cat_template = "%s '--%s ' o {%s}"
    return cat_template % (param.name, param.name,
                           ",".join([str(value) for value in param.sequence]))

def build_constant(param):
    constant_template = "%s '--%s ' c (%s)"
    return constant_template % (param.name, param.name, param.value)


def build_continuous(param):
    if type(param) in (NormalIntegerHyperparameter,
                       NormalFloatHyperparameter):
        param = param.to_uniform()

    float_template = "%s '--%s ' r (%f, %f)"
    int_template = "%s '--%s ' i (%d, %d)"

    if param.log:
        lower = np.log(param.lower)
        upper = np.log(param.upper)
        # default = np.log(param.default)

    else:
        lower = param.lower
        upper = param.upper
        # default = param.default

    # if param.q is not None
    #     q_prefix = "Q%d_" % (int(param.q),)
    # else:
    #     q_prefix = ""

    if isinstance(param, IntegerHyperparameter):
        return int_template % (param.name, param.name, int(lower),
                               int(upper))
    else:
        return float_template % (param.name, param.name, float(lower),
                                 float(upper))


def build_condition(condition):
    if not isinstance(condition, ConditionComponent):
        raise TypeError("build_condition must be called with an instance of "
                        "'%s', got '%s'" %
                        (ConditionComponent, type(condition)))

    # Now handle the conditions IRACE can handle
    in_template = "%s %%in%% %s(%s)"
    less_template = "%s < %s"
    greater_template = "%s > %s"
    notequal_template = "%s != %s"
    equal_template = "%s==%s"

    full_condition = ''
    child = ''
    conditions = [condition] # if not a conjunction this makes condition iteratable

    # Check if IRACE can handle the condition
    if isinstance(condition, OrConjunction):
        logic = " || "
        conditions = condition.components

    elif isinstance(condition, AndConjunction):
        logic = " && "
        conditions = condition.components

    else:
        logic = None

    for clause in conditions:
        child = clause.child.name
        # Findout type of parent
        if isinstance(clause.parent, UniformIntegerHyperparameter) or isinstance(clause.parent,
                                                                               NormalIntegerHyperparameter):
            pType = 'i'
        elif isinstance(clause.parent, UniformFloatHyperparameter) or isinstance(clause.parent,
                                                                                 NormalFloatHyperparameter):
            pType = 'r'
        elif isinstance(clause.parent, CategoricalHyperparameter):
            pType = 'c'
        elif isinstance(clause.parent, OrdinalHyperparameter):
            pType = 'o'
        else:
            raise TypeError("Parent Type of Condition is unknown")

        # check if  parent type is log integer, if it is then convert it to log manually
        # this conversion of log is needed because irace doesnt natively support params of type log
        if pType == "i" or pType == "r":
            if clause.parent.log:
                # unlike other condition types which have variable called 'value'
                # the InCondition has array called 'values'
                if hasattr(clause, 'values'):
                    clause.values = np.log(clause.values)
                else:
                    clause.value = np.log(clause.value)

        if full_condition != '':
            full_condition += logic

        if isinstance(clause, NotEqualsCondition):
            full_condition += notequal_template % (clause.parent.name, clause.value)

        elif isinstance(clause, InCondition):
            full_condition += in_template % (clause.parent.name, pType, ",".join(clause.values))

        elif isinstance(clause, EqualsCondition):
            full_condition += equal_template % (clause.parent.name, clause.value)

        elif isinstance(clause, LessThanCondition):
            full_condition += less_template % (clause.parent.name, clause.value)

        elif isinstance(clause, GreaterThanCondition):
            full_condition += greater_template % (clause.parent.name, clause.value)

    return child + ' | ' + full_condition



def build_forbidden(clause):
    if not isinstance(clause, AbstractForbiddenComponent):
        raise TypeError("build_forbidden must be called with an instance of "
                        "'%s', got '%s'" %
                        (AbstractForbiddenComponent, type(clause)))

    if not isinstance(clause, (ForbiddenEqualsClause, ForbiddenAndConjunction)):
        raise NotImplementedError("IRACE cannot handle '%s' of type %s" %
                                  str(clause), (type(clause)))

    retval = io.StringIO()
    retval.write("(")
    # Really simple because everything is an AND-conjunction of equals
    # conditions
    dlcs = clause.get_descendant_literal_clauses()
    for dlc in dlcs:
        if retval.tell() > 1:
            retval.write(" && ")
        retval.write("%s==%s" % (dlc.hyperparameter.name, dlc.value))
    retval.write(")")
    retval.seek(0)
    return retval.getvalue()


def write(configuration_space):
    if not isinstance(configuration_space, ConfigurationSpace):
        raise TypeError("irace.write expects an instance of %s, "
                        "you provided '%s'" % (ConfigurationSpace,
                                               type(configuration_space)))

    param_lines = io.StringIO()
    condition_lines = io.StringIO()
    forbidden_lines = []
    for hyperparameter in configuration_space.get_hyperparameters():
        # Check if the hyperparameter names are valid IRACE names!
        try:
            pp_param_name.parseString(hyperparameter.name)
        except pyparsing.ParseException:
            raise ValueError(
                "Illegal hyperparameter name for IRACE: %s" % hyperparameter.name)

        # First build params
        if param_lines.tell() > 0:
            param_lines.write("\n")
        if isinstance(hyperparameter, NumericalHyperparameter):
            # print "building countinuous param"
            param_lines.write(build_continuous(hyperparameter))

        elif isinstance(hyperparameter, CategoricalHyperparameter):
            # print "building categorical param"
            param_lines.write(build_categorical(hyperparameter))

        elif isinstance(hyperparameter, Constant):
            # print "building constant param"
            param_lines.write(build_constant(hyperparameter))

        elif isinstance(hyperparameter, OrdinalHyperparameter):
            # print "building constant param"
            param_lines.write(build_ordinal(hyperparameter))

        else:
            raise TypeError("Unknown type: %s (%s)" % (
                type(hyperparameter), hyperparameter))

    for condition in configuration_space.get_conditions():
        if condition_lines.tell() > 0:
            condition_lines.write("\n")
        condition_lines.write(build_condition(condition))

    for forbidden_clause in configuration_space.get_forbiddens():
        # Convert in-statement into two or more equals statements
        dlcs = forbidden_clause.get_descendant_literal_clauses()
        # First, get all in statements and convert them to equal statements
        in_statements = []
        other_statements = []
        for dlc in dlcs:
            if isinstance(dlc, MultipleValueForbiddenClause):
                if not isinstance(dlc, ForbiddenInClause):
                    raise ValueError("IRACE cannot handle this forbidden "
                                     "clause: %s" % dlc)
                in_statements.append(
                    [ForbiddenEqualsClause(dlc.hyperparameter, value)
                     for value in dlc.values])
            else:
                other_statements.append(dlc)
        # Second, create the product of all elements in the IN statements,
        # create a ForbiddenAnd and add all ForbiddenEquals
        if len(in_statements) > 0:
            for i, p in enumerate(product(*in_statements)):
                all_forbidden_clauses = list(p) + other_statements
                f = ForbiddenAndConjunction(*all_forbidden_clauses)
                forbidden_lines.append(build_forbidden(f))
        else:
            forbidden_lines.append(build_forbidden(forbidden_clause))

    # Add conditions: first convert param_lines to array then search first part of condition in that array
    # if found append second part of condition to that array part
    splitted_params = param_lines.getvalue().split("\n")
    if condition_lines.tell() > 0:
        condition_lines.seek(0)
        param_lines.write("\n\n")
        for line in condition_lines:
            param_lines.write(line)
            t = filter(lambda x: line.split(" ")[0] in x, splitted_params)
            index = splitted_params.index(next(t))
            splitted_params[index] = splitted_params[index] + "  ".join(line.split(" ")[1:])


    for i, j in enumerate(splitted_params):
        if j[-1] != "\n":
            splitted_params[i] += "\n"

    forbidden_lines_write = io.StringIO()
    if len(forbidden_lines) > 0:
        for forbidden in forbidden_lines:
            forbidden_lines_write.write(forbidden + '\n')

    output_fh = open('forbidden.txt', 'w')
    output_fh.write(forbidden_lines_write.getvalue())
    output_fh.close()

    # overwrite param_lines with split_params which contains lines with conditions
    param_lines = io.StringIO()
    for l in splitted_params:
        param_lines.write(l)
    return param_lines.getvalue()



# Copyright (c) 2014-2016, ConfigSpace developers
# Matthias Feurer
# Katharina Eggensperger
# and others (see commit history).
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import unittest

import ConfigSpace
import ConfigSpace.util
import ConfigSpace.read_and_write.pcs as pcs_parser
import ConfigSpace.read_and_write.pcs_new as pcs_new_parser


class ExampleSearchSpacesTest(unittest.TestCase):
    pass


def generate(configuration_space_path):
    def run_test(self):
        try:
            with open(configuration_space_path) as fh:
                cs = pcs_parser.read(fh)
        except:
            with open(configuration_space_path) as fh:
                cs = pcs_new_parser.read(fh)

        default = cs.get_default_configuration()
        cs._check_configuration_rigorous(default)
        for i in range(10):
            neighborhood = ConfigSpace.util.get_one_exchange_neighbourhood(
                default, seed=i)

            for shuffle, n in enumerate(neighborhood):
                n.is_valid_configuration()
                cs._check_configuration_rigorous(n)
                if shuffle == 10:
                    break

        # Sample a little bit
        for i in range(10):
            cs.seed(i)
            configurations = cs.sample_configuration(size=5)
            for j, c in enumerate(configurations):
                c.is_valid_configuration()
                cs._check_configuration_rigorous(c)
                neighborhood = ConfigSpace.util.get_one_exchange_neighbourhood(
                    c, seed=i)

                for shuffle, n in enumerate(neighborhood):
                    n.is_valid_configuration()
                    cs._check_configuration_rigorous(n)
                    if shuffle == 20:
                        break
    return run_test


this_file = os.path.abspath(__file__)
this_directory = os.path.dirname(this_file)
configuration_space_path = os.path.join(this_directory,
                                        "..", "test_searchspaces")
configuration_space_path = os.path.abspath(configuration_space_path)
pcs_files = sorted(os.listdir(configuration_space_path))

for pcs_file in pcs_files:
    if '.pcs' in pcs_file:
        full_path = os.path.join(configuration_space_path, pcs_file)
        setattr(ExampleSearchSpacesTest, 'test_%s' % pcs_file.replace('.', '_'),
                generate(full_path))

if __name__ == '__main__':
    suite = unittest.TestSuite()
    suite.addTest(ExampleSearchSpacesTest(
        methodName='test_auto-sklearn_2017_04_pcs'))
    runner = unittest.TextTestRunner().run(suite)



import numpy as np
cimport numpy as np

# We now need to fix a datatype for our arrays. I've used the variable
# DTYPE for this, which is assigned to the usual NumPy runtime
# type info object.
DTYPE = np.float
# "ctypedef" assigns a corresponding compile-time type to DTYPE_t. For
# every type in the numpy module there's a corresponding compile-time
# type with a _t-suffix.
ctypedef np.float_t DTYPE_t


cdef class Hyperparameter(object):
    cdef public str name
    cdef public default_value
    cdef public DTYPE_t normalized_default_value
    cdef public dict meta

    cpdef int compare_vector(self, DTYPE_t value, DTYPE_t value2)
    cpdef bint is_legal_vector(self, DTYPE_t value)
